#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""Global constants."""

ARCHES_ALL = ["amd64", "arm64", "armhf", "armel", "i386", "powerpc", "ppc64el", "s390x", "riscv64"]
ARCHES_PPA_DEFAULT = ["amd64", "i386"]
ARCHES_PPA_ALL = ["amd64", "arm64", "armhf", "i386", "ppc64el", "s390x", "riscv64"]
ARCHES_PPA_EXTRA = []
ARCHES_PPA_LEGACY = list(set(ARCHES_ALL) - set(ARCHES_PPA_ALL + ARCHES_PPA_EXTRA))

ARCHES_AUTOPKGTEST = ["amd64", "arm64", "armhf", "i386", "ppc64el", "s390x", "riscv64"]

CREDENTIALS_FILENAME_DEFAULT = "credentials.oauth"

URL_LPAPI = "https://api.launchpad.net/devel"
URL_AUTOPKGTEST = "https://autopkgtest.ubuntu.com"

DISTRO_UBUNTU_COMPONENTS = ['main', 'restricted', 'universe', 'multiverse', 'partner']

DISTRO_UBUNTU_POCKETS = ['release', 'security', 'proposed', 'updates', 'backports']
DISTRO_UBUNTU_POCKETS_UPDATES = ['release', 'security', 'updates']

LOCAL_REPOSITORY_PATH = "/tmp/ubuntu"
LOCAL_REPOSITORY_MIRRORING_DIRECTIONS = f"""
Tip: You can generate (and refresh) a dists-only mirror thusly:
  $ mkdir {LOCAL_REPOSITORY_PATH}
  $ rsync -va \\
      --exclude={{'*/installer*','*/i18n/*','*/uefi/*','*/Contents*','*/by-hash/*','*tar.gz'}} \\
      rsync://archive.ubuntu.com/ubuntu/dists {LOCAL_REPOSITORY_PATH}

It's recommended to run the rsync command as a cronjob to keep your
repository up to date as often as desired.
"""



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""An individual autopkgtest run."""

import json
from functools import lru_cache
from typing import Iterator
import urllib

from .constants import URL_AUTOPKGTEST


class Job:
    """An individual autopkgtest run that has not yet completed.

    A Job will correspond to one Result object once it has completed.
    """
    def __init__(self, number, submit_time, source_package, series, arch,
                 triggers=None, ppas=None):
        """Initialize a new Job object.

        :param str number: Position within the waiting queue.
        :param str submit_time: Timestamp when job was submitted.
        :param str source_package: Source package containing the DEP8 tests to run.
        :param str series: Codename of the Ubuntu release to run tests on.
        :param str arch: Hardware architecture type to run tests on.
        :param list[str] triggers: List of package/version triggers for the job.
        :param list[str] ppas: List of PPAs to enable.
        """
        self.number = number
        self.submit_time = submit_time
        self.source_package = source_package
        self.series = series
        self.arch = arch
        self.triggers = triggers or []
        self.ppas = ppas or []

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'source_package={self.source_package!r}, '
                f'series={self.series!r}, '
                f'arch={self.arch!r}'
                f')')

    def __str__(self) -> str:
        """Return a human-readable summary of the object.

        :rtype: str
        :returns: Printable summary of the object.
        """
        return f"{self.source_package} {self.series} ({self.arch})"

    @lru_cache
    def to_dict(self):
        return {
            'number': self.number,
            'submit_time': self.submit_time,
            'source_package_name': self.source_package,
            'series': self.series,
            'arch': self.series,
            'triggers': self.triggers,
            'ppas': self.ppas
        }

    @property
    def request_url(self) -> str:
        """Render URL for requesting the testing run be started.

        :rtype: str
        :returns: Full URL for invoking the test.
        """
        parameter_str = urllib.parse.urlencode({
            'release': self.series,
            'arch': self.arch,
            'package': self.source_package})
        for trigger in self.triggers:
            parameter_str += "&" + urllib.parse.urlencode({"trigger": trigger})
        return f"{URL_AUTOPKGTEST}/request.cgi?{parameter_str}"


def get_running(response, releases=None, sources=None, ppa=None) -> Iterator[Job]:
    """Return iterator currently running autopkgtests for given criteria.

    Filters the list of running autopkgtest jobs by the given series
    and/or ppa names, returning an iterator with matching results as Job
    objects.  If series and ppa are not provided, then returns all
    results; if one or the other is provided, provides all available
    results for that series or ppa.

    :param HTTPResponse response: Context manager; the response from urlopen()
    :param List[str] releases: The Ubuntu series codename(s), or None.
    :param List[str] sources: Only retrieve results for these
        source packages, or all if blank or None.
    :param str ppa: The PPA address criteria, or None.
    :rtype: Iterator[Job]
    :returns: Currently running jobs, if any, or an empty list on error
    """
    for pkg, jobs in json.loads(response.read().decode('utf-8') or '{}').items():
        if sources and (pkg not in sources):
            continue
        for handle in jobs:
            for codename in jobs[handle]:
                for arch, jobinfo in jobs[handle][codename].items():
                    triggers = jobinfo[0].get('triggers', None)
                    ppas = jobinfo[0].get('ppas', None)
                    submit_time = jobinfo[1]
                    job = Job(0, submit_time, pkg, codename, arch, triggers, ppas)
                    if releases and (job.series not in releases):
                        continue
                    if ppa and (ppa not in job.ppas):
                        continue
                    yield job


def get_waiting(response, releases=None, sources=None, ppa=None) -> Iterator[Job]:
    """Return iterator of queued autopkgtests for given criteria.

    Filters the list of autopkgtest jobs waiting for execution by the
    given series and/or ppa names, returning an iterator with matching
    results as Job objects.  If series and ppa are not provided, then
    returns all results; if one or the other is provided, provides all
    available results for that series or ppa.

    :param HTTPResponse response: Context manager; the response from urlopen()
    :param List[str] releases: The Ubuntu series codename(s), or None.
    :param List[str] sources: Only retrieve results for these
        source packages, or all if blank or None.
    :param str ppa: The PPA address criteria, or None.
    :rtype: Iterator[Job]
    :returns: Currently waiting jobs, if any, or an empty list on error
    """
    for _, queue in json.loads(response.read().decode('utf-8') or '{}').items():
        for codename in queue:
            for arch in queue[codename]:
                n = 0
                for key in queue[codename][arch]:
                    if key == 'private job':
                        continue
                    (pkg, json_data) = key.split(maxsplit=1)
                    if sources and (pkg not in sources):
                        continue
                    jobinfo = json.loads(json_data)
                    n += 1
                    triggers = jobinfo.get('triggers', None)
                    ppas = jobinfo.get('ppas', None)
                    job = Job(n, None, pkg, codename, arch, triggers, ppas)
                    if releases and (job.series not in releases):
                        continue
                    if ppa and (ppa not in job.ppas):
                        continue
                    yield job


def show_running(jobs):
    """Print the active (running and waiting) tests."""
    rformat = "%-8s %-20s %-8s %-8s %-25s %s"

    n = 0
    for n, e in enumerate(jobs, start=1):
        if n == 1:
            print("* Running:")
            t_str = str(e.submit_time)
            ppa_str = ','.join(e.ppas)
            trig_str = ','.join(e.triggers)
            print("  # " + rformat % ("time", "pkg", "release", "arch", "ppa", "trigger"))
        print("  - " + rformat % (t_str, e.source_package, e.series, e.arch, ppa_str, trig_str))
    if n == 0:
        print("* Running: (none)")


def show_waiting(jobs):
    """Print the active (running and waiting) tests."""
    rformat = "%-8s %-40s %-8s %-8s %-40s %s"

    n = 0
    for n, e in enumerate(jobs, start=1):
        if n == 1:
            print("* Waiting:")
            print("  # " + rformat % ("Q-num", "pkg", "release", "arch", "ppa", "trigger"))

        ppa_str = ','.join(e.ppas)
        trig_str = ','.join(e.triggers)
        print("  - " + rformat % (e.number, e.source_package, e.series, e.arch, ppa_str, trig_str))
    if n == 0:
        print("* Waiting: (none)")


if __name__ == "__main__":
    import os
    from urllib.request import urlopen

    print('############################')
    print('### Job class smoke test ###')
    print('############################')
    print()

    root_dir = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
    jobinfo = {
        'triggers': ['a/1', 'b/2.1', 'c/3.2.1'],
        'ppas': ['ppa:me/myppa']
    }
    job_1 = Job(
        number=0,
        submit_time='time',
        source_package='my-package',
        series='kinetic',
        arch='amd64',
        triggers=jobinfo.get('triggers', None),
        ppas=jobinfo.get('ppas', None)
    )
    print(job_1)
    print(f"triggers:     {job_1.triggers}")
    print(f"ppas:         {job_1.ppas}")
    print(f"request_url:  {job_1.request_url}")
    print()

    # pylint: disable-next=invalid-name
    ppa = "bryce/dovecot-merge-v1e2.3.19.1adfsg1-2"

    print("running:")
    response = urlopen(f"file://{root_dir}/tests/data/running-20220822.json")
    for job in get_running(response, releases=['kinetic'], sources=None, ppa=ppa):
        print(job)
    print()

    print("waiting:")
    response = urlopen(f"file://{root_dir}/tests/data/queues-20220822.json")
    for job in get_waiting(response, releases=['kinetic'], sources=None, ppa=ppa):
        print(job)
    print()

    print("Object Dump")
    print("-----------")
    print(json.dumps(job_1.to_dict(), indent=4))
    print()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2022 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the text helper routines."""

import os
import sys

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

import ppa.text


@pytest.mark.parametrize('input, expected', [
    ('true', True),
    ('t', True),
    ('yes', True),
    ('y', True),
    ('1', True),
    ('false', False),
    ('f', False),
    ('no', False),
    ('n', False),
    ('0', False),
    ('', False),
    (1, True),
    (0, False),
    (1.0, True),
    (0.0, False),
    ((), False),
    ((1,), True),
    (None, False),
    (object(), True),
])
def test_to_bool(input, expected):
    """Check that the given input produces the expected true/false result.

    :param * input: Any available type to be converted to boolean.
    :param bool expected: The True or False result to expect.
    """
    assert ppa.text.to_bool(input) == expected


def test_ansi_hyperlink():
    """Check that text can be linked with a url."""
    assert ppa.text.ansi_hyperlink("xxx", "yyy") == "\u001b]8;;xxx\u001b\\yyy\u001b]8;;\u001b\\"



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2023 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests helper routines for handling dict objects."""

import os
import sys
import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.dict import unpack_to_dict


@pytest.mark.parametrize('text, expected', [
    ('a',                 {'a': ''}),
    ('a=1',               {'a': '1'}),
    ('a=1:2',             {'a': '1:2'}),
    ('a = 1',             {'a': '1'}),
    ('a:x',               {'a': ''}),
    ('a:x:',              {'a': ''}),
    ('a:x:y',             {'a': ''}),
    ('a:x=1',             {'a': '1'}),
    ('a:x = 1',           {'a': '1'}),
    ('a=x=1',             {'a': 'x=1'}),
    ('a : x=1',           {'a': '1'}),
    ('a,b',               {'a': '', 'b': ''}),
    ('a, b',              {'a': '', 'b': ''}),
    ('a,b=1.2.3,c:x=4',   {'a': '', 'b': '1.2.3', 'c': '4'}),
    ('a, b=1.2.3, c:x=4', {'a': '', 'b': '1.2.3', 'c': '4'}),
    ('a, b|c',            {'a': '', ('b', 'c'): {'b': '', 'c': ''}}),
    ('a, b=1|c=2',        {'a': '', ('b', 'c'): {'b': '1', 'c': '2'}}),
    ('a, b=1|c:x=2, d=3', {'a': '', ('b', 'c'): {'b': '1', 'c': '2'}, 'd': '3'}),
])
def test_unpack_to_dict(text, expected):
    """Checks the unpack_to_dict() routine's string unpacking."""
    result = unpack_to_dict(text)

    assert result
    assert isinstance(result, dict)
    assert result == expected


@pytest.mark.parametrize('text, expected_exception', [
    (None, ValueError),
    ('',   ValueError),
    (',',  ValueError),
    (',z', ValueError),
    (':',  ValueError),
    (':z', ValueError),
    ('=',  ValueError),
    ('=z', ValueError),
])
def test_unpack_to_dict_error(text, expected_exception):
    """Checks the unpack_to_dict() routine's string unpacking."""
    with pytest.raises(expected_exception):
        unpack_to_dict(text)


@pytest.mark.parametrize('text, key_cut, sep, expected', [
    ('a:x=1', ':',  '=',  {'a': '1'}),

    ('a.x=1', '.',  '=',  {'a': '1'}),
    ('a+x=1', '+',  '=',  {'a': '1'}),
    ('a-x=1', '-',  '=',  {'a': '1'}),
    ('a~x=1', '~',  '=',  {'a': '1'}),
    ('a_x=1', '_',  '=',  {'a': '1'}),
    ('a!x=1', '!',  '=',  {'a': '1'}),
    ('a;x=1', ';',  '=',  {'a': '1'}),
    ('a/x=1', '/',  '=',  {'a': '1'}),

    ('a:x.1', ':',  '.',  {'a': '1'}),
    ('a:x+1', ':',  '+',  {'a': '1'}),
    ('a:x-1', ':',  '-',  {'a': '1'}),
    ('a:x~1', ':',  '~',  {'a': '1'}),
    ('a:x_1', ':',  '_',  {'a': '1'}),
    ('a:x!1', ':',  '!',  {'a': '1'}),
    ('a:x;1', ':',  ';',  {'a': '1'}),
    ('a:x/1', ':',  '/',  {'a': '1'}),

    # Spaces are allowed as separators
    ('a 1', ':', ' ', {'a': '1'}),
    ('a 1, b, c 3', ':', ' ', {'a': '1', 'b': '', 'c': '3'}),

])
def test_unpack_to_dict_parameters(text, sep, key_cut, expected):
    """Checks the unpack_to_dict() routine's string unpacking."""
    result = unpack_to_dict(text, key_sep=sep, key_cut=key_cut)

    assert result
    assert isinstance(result, dict)
    assert result == expected


@pytest.mark.parametrize('text, key_cut, sep, expected_exception', [
    # Commas are reserved as the item separator
    ('a:x=1', ',',  '=',  ValueError),
    ('a:x=1', ':',  ',',  ValueError),
    ('a:x=1', ',',  ',',  ValueError),

    # key_cut and sep have to be different
    ('a:x=1', ':',  ':', ValueError),
    ('a:x=1', '=',  '=', ValueError),
    ('a:x=1', '.',  '.', ValueError),
])
def test_unpack_to_dict_parameters_error(text, sep, key_cut, expected_exception):
    """Checks the unpack_to_dict() error handling, with invalid parameters."""
    with pytest.raises(expected_exception):
        unpack_to_dict(text, key_sep=sep, key_cut=key_cut)



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2022 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests for utilities used to read and write data externally."""

import os
import sys
import urllib

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.io import open_url


def test_open_url(tmp_path):
    """Checks that the open_url() object reads from a valid URL.

    :param fixture tmp_path: Temp dir.
    """
    f = tmp_path / "open_url.txt"
    f.write_text("abcde")

    request = open_url(f"file://{f}")
    assert request
    assert isinstance(request, urllib.response.addinfourl)
    assert request.read().decode() == 'abcde'



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""Top-level code for analyzing an Ubuntu Apt repository."""

import os.path
from functools import lru_cache
from typing import Dict

from .suite import Suite
from .constants import (
    LOCAL_REPOSITORY_PATH,
    LOCAL_REPOSITORY_MIRRORING_DIRECTIONS
)


class Repository:
    """Top-level class for analyzing an Ubuntu Apt repository.

    This class and its children serves as a wrapper to a local mirror of
    a Ubuntu Apt repository.  Repository provides an entry interface to
    the Apt data, allowing lazy lookup of the package data in various ways.
    """

    def __init__(self, cache_dir: str):
        """Initialize a new Repository object for an Apt cache.

        :param str cache_dir: The path to the top level of the local Apt mirror.
        """
        if not cache_dir:
            raise ValueError("undefined cache_dir.")
        if not os.path.isdir(cache_dir):
            raise FileNotFoundError(f"could not find cache dir '{cache_dir}'")

        self.cache_dir = cache_dir

    def __repr__(self):
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'cache_dir={self.cache_dir!r})')

    @property
    @lru_cache
    def suites(self) -> Dict[str, Suite]:
        """The release pockets available in this repository.

        :returns: Mapping of Ubuntu release pockets to corresponding Suite objects.
        :rtype: Dict[str, Suite]
        """
        return {
            suite_name: Suite(suite_name, os.path.join(self.cache_dir, suite_name))
            for suite_name
            in os.listdir(self.cache_dir)
            if os.path.isdir(os.path.join(self.cache_dir, suite_name))
        }

    def get_suite(self, series_codename: str, pocket: str) -> Suite:
        """Retrieve a Suite object by its codename and pocket.

        The pocket name 'release' is treated as synonymous with '',
        since Apt stores the release suite without the pocket name.  In
        other words, 'lunar-release' is the same as 'lunar'.

        :param str series_codename: The Ubuntu release textual name.
        :param str pocket: The name of the pocket ('proposed', etc.)
        :returns: Corresponding Suite object.
        :rtype: Suite
        """
        if pocket == "release":
            release_pocket = series_codename
        else:
            release_pocket = f"{series_codename}-{pocket}"
        return self.suites.get(release_pocket, None)


if __name__ == "__main__":
    import sys
    from pprint import PrettyPrinter
    pp = PrettyPrinter(indent=4)

    from .debug import error

    print('#########################')
    print('## PpaGroup smoke test ##')
    print('#########################')

    local_dists_path = os.path.join(LOCAL_REPOSITORY_PATH, "dists")
    if not os.path.exists(local_dists_path):
        error(f'Missing checkout for smoketest\n{LOCAL_REPOSITORY_MIRRORING_DIRECTIONS}')
        sys.exit(1)
    repository = Repository(cache_dir=local_dists_path)
    for suite in repository.suites.values():
        print(suite)
        print(f'  series:      {suite.series_codename}')
        print(f'  components:  {", ".join(suite.components)}')



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2022-2024 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the Trigger class for managing Autopkgtest trigger URLs."""

import os
import sys

import json
import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.trigger import Trigger, get_triggers, show_triggers
from ppa.source_publication import SourcePublication


def test_object():
    """Checks that Trigger objects can be instantiated."""
    trigger = Trigger(SourcePublication('x', 'x', 'x'), 'x')
    assert trigger


@pytest.mark.parametrize('pub, arch, test_pkg, expected_repr', [
    (
        SourcePublication('pkg', 'ver', 'ser'),
        'arch',
        'tst',
        (
            "Trigger(source_publication=SourcePublication("
            "package='pkg', version='ver', series='ser', archive=None, status=None), "
            "architecture='arch', "
            "test_package='tst')"
        )
    ),
    (
        SourcePublication('pkg', 'ver', 'ser'),
        'arch',
        None,
        (
            "Trigger(source_publication=SourcePublication("
            "package='pkg', version='ver', series='ser', archive=None, status=None), "
            "architecture='arch', "
            "test_package='pkg')"
        )
    ),
])
def test_repr(pub, arch, test_pkg, expected_repr):
    """Checks Trigger object representation."""
    trigger = Trigger(pub, arch, test_pkg)
    assert repr(trigger) == expected_repr


@pytest.mark.parametrize('pub, arch, test_pkg, expected_str', [
    (
        SourcePublication('pkg', 'ver', 'ser'),
        'arch',
        None,
        'pkg: ser/pkg/ver [arch]'
    ),
    (
        SourcePublication('pkg', 'ver', 'ser', 'ppa'),
        'arch',
        None,
        'pkg: ser/pkg/ver (ppa) [arch]'
    ),
    (
        SourcePublication('dovecot', '1:2.3.19.1+dfsg1-2ubuntu2', 'kinetic', None),
        'armhf',
        'apache2',
        'apache2: kinetic/dovecot/1:2.3.19.1+dfsg1-2ubuntu2 [armhf]'
    ),
])
def test_str(pub, arch, test_pkg, expected_str):
    """Checks Trigger object textual presentation."""
    trigger = Trigger(pub, arch, test_pkg)
    assert f"{trigger}" == expected_str


def test_to_dict():
    """Checks Trigger object structural representation."""
    trigger = Trigger(SourcePublication('x', 'x', 'x', 'x'), 'x', 'x')
    expected_keys = [
        'architecture', 'source_publication', 'test_package',
    ]
    expected_types = [str, dict]

    d = trigger.to_dict()
    assert isinstance(d, dict), f"type of d is {type(d)} not dict"

    # Verify expected keys are present
    assert sorted(d.keys()) == sorted(expected_keys)

    # Verify values are within set of expected types
    for k, v in d.items():
        assert type(v) in expected_types, f"'{k}={v}' is unexpected type {type(v)}"

    # Verify full dict can be written as JSON
    try:
        assert json.dumps(d)
    except UnicodeDecodeError as e:
        assert False, f"Wrong UTF codec detected: {e}"
    except json.JSONDecodeError as e:
        assert False, f"JSON decoding error: {e.msg}, {e.doc}, {e.pos}"


@pytest.mark.parametrize('pub, arch, test_pkg, expected', [
    (
        SourcePublication('pkg', 'ver', 'ser'),
        'arch',
        None,
        "/request.cgi?release=ser&package=pkg&arch=arch&trigger=pkg%2Fver"
    ),
    (
        SourcePublication('p', 'v', 's', 'ppa:a/b'),
        'a',
        None,
        "/request.cgi?release=s&package=p&arch=a&trigger=p%2Fv&ppa=ppa%3Aa%2Fb"
    ),
    (
        SourcePublication('p', '1.2+git345', 's'),
        'a',
        None,
        "/request.cgi?release=s&package=p&arch=a&trigger=p%2F1.2%2Bgit345"
    ),
    (
        SourcePublication('apache2', '2.4', 'kinetic'),
        'a',
        'pkg',
        "/request.cgi?release=kinetic&package=pkg&arch=a&trigger=apache2%2F2.4"
    )
])
def test_action_url(pub, arch, test_pkg, expected):
    """Checks that Trigger objects generate valid autopkgtest action urls."""
    trigger = Trigger(pub, arch, test_pkg)
    assert expected in trigger.action_url


@pytest.mark.parametrize('params, expected', [
    (
        ['pkg', 'ver', None, 'ser', [], None],
        [],
    ),
    (
        ['pkg', 'ver', None, 'ser', ['arch'], 'tst'],
        [
            Trigger(SourcePublication('pkg', 'ver', 'ser'), 'arch', 'tst'),
        ],
    ),
    (
        ['pkg', 'ver', 'ppa', 'ser', ['ax', 'ay', 'az'], 'tst'],
        [
            Trigger(SourcePublication('pkg', 'ver', 'ser', 'ppa'), 'ax', 'tst'),
            Trigger(SourcePublication('pkg', 'ver', 'ser', 'ppa'), 'ay', 'tst'),
            Trigger(SourcePublication('pkg', 'ver', 'ser', 'ppa'), 'az', 'tst'),
        ],
    ),
])
def test_get_triggers(params, expected):
    """Checks that Trigger objects get generated properly from inputs."""
    for trigger in get_triggers(*params):
        assert repr(trigger) in [repr(t) for t in expected]


@pytest.mark.parametrize('triggers, params, expected_in_stdout', [
    (
        # Basic function parameters, no triggers
        [],
        {},
        [""]
    ),
    (
        # Specified trigger (clickable)
        [Trigger(SourcePublication('p', 'v', 's', 'ppa'), 'a')],
        {'show_trigger_urls': False},
        [
            "&trigger=p%2Fv&ppa=ppa\x1b\\Trigger basic @a♻️ \x1b]8;;\x1b",
            "&trigger=p%2Fv&ppa=ppa&all-proposed=1\x1b\\Trigger all-proposed @a💍\x1b]8;;\x1b\\\n"
        ]
    ),
    (
        # Specified trigger (display plain URLs)
        [Trigger(SourcePublication('pkg', '123', 'x-series', 'y-ppa'), 'i386')],
        {'show_trigger_urls': True},
        [
            "&trigger=pkg%2F123&ppa=y-ppa ♻️ \n",
            "&trigger=pkg%2F123&ppa=y-ppa&all-proposed=1 💍\n"
        ]
    ),
    (
        # Display names of packages in trigger lines
        [Trigger(SourcePublication('pkg', '123', 'x-series'), 'i386')],
        {'show_trigger_names': True},
        ["Trigger basic pkg@i386♻️ ", "Trigger all-proposed pkg@i386💍"]
    ),
    (
        # Omit package names if specified
        [Trigger(SourcePublication('pkg', '123', 'x-series'), 'i386')],
        {'show_trigger_names': False},
        ["Trigger basic @i386", "Trigger all-proposed @i386"]
    ),
    (
        # Display names of packages in trigger lines when trigger URLs are shown
        [Trigger(SourcePublication('pkg', '123', 'x-series'), 'i386')],
        {'show_trigger_urls': True, 'show_trigger_names': True},
        ["pkg@i386: https:", "trigger=pkg%2F123"]
    ),
    (
        # Multiple triggers
        [
            Trigger(SourcePublication('pkg', '123', 'x-series'), 'i386'),
            Trigger(SourcePublication('lib', '321', 'x-series'), 'i386')
        ],
        {'show_trigger_urls': True},
        ["trigger=pkg%2F123", "trigger=lib%2F321"]
    ),
])
def test_show_triggers(capfd, triggers, params, expected_in_stdout):
    """Check that trigger strings are included in generated output."""
    params.setdefault('triggers', [])
    for t in triggers:
        params['triggers'].append(t)
    show_triggers(**params)
    out, _ = capfd.readouterr()
    print(out)
    for text in expected_in_stdout:
        assert text in out



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2023 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the Repository class as an interface to an Apt repository."""

import os
import sys

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.repository import Repository
from ppa.suite import Suite


def test_object(tmp_path):
    """Checks that Repository objects can be instantiated."""
    repository = Repository(tmp_path)
    assert repository


def test_suites(tmp_path):
    """Checks getting all suites from the repository."""
    suites = ['a', 'b', 'b-0', 'b-1']
    for suite in suites:
        suite_dir = tmp_path / suite
        suite_dir.mkdir()

    repository = Repository(tmp_path)
    assert sorted(repository.suites.keys()) == sorted(suites)
    for suite in repository.suites.values():
        assert isinstance(suite, Suite)


def test_get_suite(tmp_path):
    """Checks getting a specific suite from the repository."""
    suites = ['a', 'b', 'b-0', 'b-1']
    for suite in suites:
        suite_dir = tmp_path / suite
        suite_dir.mkdir()

    repository = Repository(tmp_path)
    suite = repository.get_suite('b', '1')
    assert suite
    assert str(suite) == 'b-1'



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2024 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests SourcePublication class as representing a package release."""

import os
import sys

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.source_publication import SourcePublication  # noqa: E402


@pytest.mark.parametrize('pub_dict', [
    ({'package': 'pkg', 'version': 'ver', 'series': 'ser'}),
    ({'package': 'pkg2.0-abc', 'version': '1.0-1', 'series': 'noble'}),
    ({'package': 'dovecot', 'version': '1.2.3', 'series': 'noble', 'archive': 'ppa:me/myppa'}),
])
def test_object(pub_dict):
    """Check that SourcePublication objects can be instantiated."""
    assert SourcePublication(**pub_dict)


@pytest.mark.parametrize('pub_dict, expected_exception', [
    ({}, TypeError),
    ({'package': 'x'}, TypeError),
    ({'version': 'x'}, TypeError),
    ({'series': 'x'}, TypeError),
    ({'package': None, 'version': None, 'series': None}, ValueError),
    ({'package': None, 'version': 'x',  'series': None}, ValueError),
    ({'package': 'x',  'version': None, 'series': None}, ValueError),
    ({'package': 'x',  'version': 'x',  'series': None}, ValueError),
    ({'package': 'x',  'version': 'x',  'series': None}, ValueError),
    ({'package': None, 'version': 'x',  'series': 'x'}, ValueError),
])
def test_object_error(pub_dict, expected_exception):
    """Check that SourcePublication objects can be instantiated."""
    with pytest.raises(expected_exception):
        source_publication = SourcePublication(**pub_dict)
        assert source_publication


@pytest.mark.parametrize('pub_dict, expected_repr', [
    (
        {'package': 'pkg', 'version': 'ver', 'series': 'ser'},
        "SourcePublication(package='pkg', version='ver', series='ser', archive=None, status=None)",
    ),
    (
        {'package': 'pkg', 'version': '1.2.3', 'series': 'noble', 'archive': 'ppa:me/myppa', 'status': 'pass'},
        "SourcePublication(package='pkg', version='1.2.3', series='noble', archive='ppa:me/myppa', status='pass')",
    ),
])
def test_repr(pub_dict, expected_repr):
    """Check that SourcePublication objects can be instantiated."""
    source_publication = SourcePublication(**pub_dict)
    assert repr(source_publication) == expected_repr


@pytest.mark.parametrize('pub_dict, expected_str', [
    (
        {'package': 'pkg', 'version': 'ver', 'series': 'ser'},
        "ser/pkg/ver",
    ),
    (
        {'package': 'pkg', 'version': '1.2.3', 'series': 'noble', 'archive': 'ppa:me/myppa'},
        "noble/pkg/1.2.3 (ppa:me/myppa)",
    ),
])
def test_str(pub_dict, expected_str):
    """Check that SourcePublication objects can be instantiated."""
    source_publication = SourcePublication(**pub_dict)
    assert str(source_publication) == expected_str


@pytest.mark.parametrize('pub_dict, expected_dict', [
    (
        {'package': 'pkg', 'version': 'ver', 'series': 'ser'},
        {'package': 'pkg', 'version': 'ver', 'series': 'ser', 'archive': None, 'status': None},
    ),
    (
        {'package': 'pkg', 'version': '1.2.3', 'series': 'noble', 'archive': 'ppa:me/myppa', 'status': 'pass'},
        {'package': 'pkg', 'version': '1.2.3', 'series': 'noble', 'archive': 'ppa:me/myppa', 'status': 'pass'},
    ),
])
def test_to_dict(pub_dict, expected_dict):
    """Check that SourcePublication objects can be instantiated."""
    source_publication = SourcePublication(**pub_dict)
    assert source_publication.to_dict() == expected_dict


@pytest.mark.parametrize('pub_dict, expected_url', [
    (
        {'package': 'pkg', 'version': 'x', 'series': 'ser'},
        'https://autopkgtest.ubuntu.com/packages/p/pkg/ser'
    ),
    (
        {'package': 'libabc', 'version': 'x', 'series': 'noble'},
        'https://autopkgtest.ubuntu.com/packages/liba/libabc/noble'
    ),
])
def test_autopkgtest_history_url(pub_dict, expected_url):
    """Check that valid history URLs are generated."""
    source_publication = SourcePublication(**pub_dict)
    assert source_publication.autopkgtest_history_url == expected_url



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2022 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

'''
Routines for dealing with dictionary objects.
'''


def unpack_to_dict(text, key_cut=':', key_sep='=') -> dict:
    """Convert comma-delimited data into a dictionary.

    For each item, if @param key_sep is present split on it to get the
    key and value.  If @param key_sep is not present, then the item will
    be stored as the key with an empty string as the value.

    The key is further processed by excluding anything after @param
    key_cut.  For example, with the default values for @param key_sep
    and @param key_cut, the string "a,b=1.2.3,c:xyz=42" will unpack
    into:

        {
            'a': '',
            'b': '1.2.3',
            'c': '42',
        }

    A basic use case of this routine is to split out comma-separated
    items in a command line parameter's value string.  The @param
    key_sep parameter facilitates parsing lists of key=value items.  The
    @param key_cut provides a mechanism for filtering out unwanted
    portions of key names; for example, in "mypackage/universe=1.2.3" we
    may want to ignore the 'universe' detail.

    This routine is intended to handle parsing of Debian control fields
    sufficiently to determine the package names.  This routine is NOT
    intended to be a specification-compliant parser, and in particular
    is neither idempotent nor thorough in its parsing.  See
    https://wiki.debian.org/BuildProfileSpec for details about Debian's
    control field format.

    To support these control fields, the '|' symbol is recognized as a
    way of representing multiple alternative items via a tuple of keys
    and a dict for the value.  For example, the string 'a,b=1|c=2,d=3'
    will unpack into:

        {
            'a': '',
            ('b', 'c'): {'b': '1', 'c': '2'},
            'd': '3'
        }

    :param str text: Comma-separated textual data collection.
    :param str key_cut: Ignore anything in key after this character.
    :param str key_sep: Character used to separate an item's key from its value.
    :returns: Dictionary of data->value items
    :rtype: dict[str, str]
    """
    if not text or not key_sep or not key_cut:
        # None is not handled for any of the parameters so far.
        raise ValueError("unpack_to_dict() requires non-None values for all arguments")
    elif key_sep.strip() == ',':
        # Comma is used internally as the separator character, and
        # cannot currently be configured differently, thus it can't be
        # used by key_sep as a secondary separator.
        raise ValueError("comma is reserved and cannot be used for key_sep")
    elif not key_cut.strip() or key_cut.strip() == ',':
        # Whitespace is permitted for key_sep, but not for key_cut.
        # Comma is not allowed for key_cut, for same reason as for key_sep.
        raise ValueError("key_cut must be at least one (non-comma, non-whitespace) character")
    elif key_sep.strip() == key_cut.strip():
        # Since we're splitting on key_sep, using the same string to then split
        # key_cut would be redundant and ineffective.
        raise ValueError("key_sep and key_cut must not be the same string")

    def _split_item(item, key_sep, key_cut):
        if key_sep in item:
            key, value = item.split(key_sep, 1)
        else:
            key = item
            value = ''
        if key_cut:
            key = key.split(key_cut, 1)[0]

        # Blank value is allowed, but not key.  Neither can be None.
        if not key or value is None:
            raise ValueError
        return key.strip(), value.strip()

    dictionary = {}
    for item in text.split(','):
        if not item:
            raise ValueError

        item = item.strip()
        if '|' in item:
            # Handled items with are actually multiple items OR-ed together.
            subitems = {}
            for subitem in item.split('|'):
                subitem = subitem.strip()
                if not subitem:
                    raise ValueError("Undefined element of OR ('|') clause")
                subitems.update(dict([_split_item(subitem, key_sep, key_cut)]))

            # Store multi-values using a tuple key rather than a simple string.
            dictionary[tuple(subitems.keys())] = subitems
        else:
            # Store single-values as a simple string key.
            dictionary.update(dict([_split_item(item, key_sep, key_cut)]))
    return dictionary


if __name__ == "__main__":
    # pylint: disable=line-too-long, invalid-name

    import pprint
    pp = pprint.PrettyPrinter(indent=4)

    print('#####################')
    print('## Dict smoke test ##')
    print('#####################')
    print()

    text = "a, b=1.2.3, c:x=4"
    print(text)
    pp.pprint(unpack_to_dict(text))
    print()

    print("* Conflicts:")
    text = "binutils-mingw-w64-i686 (<< 2.23.52.20130612-1+3), zlib1g (>= 1:1.1.4)"
    print(text)
    pp.pprint(unpack_to_dict(text, key_sep=' '))
    print()

    print("* Build-Depends:")
    text = "autoconf (>= 2.64), dpkg-dev (>= 1.19.0.5), bison, flex, gettext, texinfo, dejagnu, quilt, chrpath, dwz, debugedit (>= 4.16), python3:any, file, xz-utils, lsb-release, zlib1g-dev, procps, g++-aarch64-linux-gnu [amd64 i386 x32] <!nocheck>, g++-arm-linux-gnueabi [amd64 arm64 i386 x32] <!nocheck>, g++-arm-linux-gnueabihf [amd64 arm64 i386 x32] <!nocheck>, g++-powerpc64le-linux-gnu [amd64 arm64 i386 ppc64 x32] <!nocheck>, g++-s390x-linux-gnu [amd64 arm64 i386 ppc64el x32] <!nocheck>, g++-alpha-linux-gnu [amd64 i386 x32] <!nocheck>, g++-hppa-linux-gnu [amd64 i386 x32] <!nocheck>, g++-m68k-linux-gnu [amd64 i386 x32] <!nocheck>, g++-powerpc-linux-gnu [amd64 i386 ppc64el x32] <!nocheck>, g++-powerpc64-linux-gnu [amd64 i386 x32] <!nocheck>, g++-riscv64-linux-gnu [amd64 arm64 i386 ppc64el x32] <!nocheck>, g++-sh4-linux-gnu [amd64 i386 x32] <!nocheck>, g++-sparc64-linux-gnu [amd64 i386 x32] <!nocheck>, g++-i686-linux-gnu [amd64 arm64 ppc64el x32] <!nocheck>, g++-x86-64-linux-gnu [arm64 i386 ppc64el] <!nocheck>, g++-x86-64-linux-gnux32 [amd64 arm64 i386 ppc64el] <!nocheck>"  # noqa: E501
    print(text)
    pp.pprint(unpack_to_dict(text, key_sep=' '))
    print()

    print("* Depends:")
    text = "binutils-common (= 2.38.50.20220707-1ubuntu1), libbinutils (= 2.38.50.20220707-1ubuntu1), binutils-x86-64-linux-gnu (= 2.38.50.20220707-1ubuntu1)"  # noqa: E501
    print(text)
    pp.pprint(unpack_to_dict(text, key_cut='#', key_sep=' '))
    print()

    print("* Depends:")
    text = """         adduser,
         libpam-runtime,
         lsb-base,
         openssl,
         ssl-cert,
         ucf,
         ${misc:Depends},
         ${shlibs:Depends}
         """
    print(text)
    pp.pprint(unpack_to_dict(text, key_cut='#', key_sep=' '))
    print()

    print("* Depends:")
    text = """         ${misc:Depends},
         debhelper-compat (= 13),
         dpkg-dev (>= 1.15.5),
         nginx-core (<< 5.1~) | nginx-light (<< 5.1~) | nginx-extras (<< 5.1~),
         nginx-core (>= 5) | nginx-light (>= 5) | nginx-extras (>= 5)
         """
    print(text)
    pp.pprint(unpack_to_dict(text, key_cut='#', key_sep=' '))
    print()

    print("* Depends:")
    text = """         ${misc:Depends},
         debhelper-compat (= 13),
         dpkg-dev (>= 1.15.5),
         nginx-core (<< 5.1~) | nginx-light (<< 5.1~) | nginx-extras (<< 5.1~),
         nginx-core (>= 5) | nginx-light (>= 5) | nginx-extras (>= 5)
         """
    print(text)
    pp.pprint(unpack_to_dict(text, key_sep=' ', key_cut='#'))
    print()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the Ppa class as an interface to Launchpad's Ppa API."""

import os
import sys
import time
from unittest.mock import patch

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.constants import URL_AUTOPKGTEST
from ppa.ppa import Ppa, ppa_address_split, get_ppa
from ppa.ppa_group import PpaGroup
from ppa.result import Result

from helpers import LpServiceMock


def test_object():
    """Check that PPA objects can be instantiated."""
    ppa = Ppa('test-ppa-name', 'test-owner-name', ppa_description='test-desc', service='test-svc')
    assert ppa
    assert ppa.ppa_name == 'test-ppa-name'
    assert ppa.owner_name == 'test-owner-name'
    assert ppa.ppa_description == 'test-desc'


def test_repr():
    """Check Ppa object representation."""
    ppa = Ppa('a', 'b', 'c', 'd')
    assert repr(ppa) == "Ppa(ppa_name='a', owner_name='b')"


def test_str():
    """Check Ppa object textual presentation."""
    ppa = Ppa('a', 'b', 'c', 'd')
    assert f"{ppa}" == 'b/a'


def test_address():
    """Check getting the PPA address."""
    ppa = Ppa('test', 'owner')
    assert ppa.address == "ppa:owner/test"


def test_description():
    """Check specifying a description when creating a PPA."""
    ppa = Ppa('test-ppa-name', 'test-owner-name', 'test-description')

    assert 'test-description' in ppa.ppa_description


@pytest.mark.parametrize('releases, architectures, expected_num_results, expected_num_triggers', [
    ([], [], 0, 0),
    (['x'], ['amd64'], 1, 1),
    (['x', 'y', 'z'], ['amd64'], 3, 1),
    (['x'], ['amd64', 'armhf', 'i386'], 1, 3),
    (['x', 'y', 'z'], ['amd64', 'armhf', 'i386'], 3, 3),
])
@patch('ppa.ppa.get_results')
@patch('ppa.ppa.open_url')
def test_get_autopkgtest_results(mock_open_url, mock_get_results,
                                 releases, architectures, expected_num_results, expected_num_triggers):
    """Check that autopkgtest results are retrieved and iterated correctly."""
    owner_name = 'a'
    ppa_name = 'b'

    # Bypass open_url() entirely so we don't try to retrieve anything.
    # Usually this returns a response that if valid is then passed to
    # get_results(), but since we'll be patching get_results(), all that
    # matters is that validity check passes.  We can do that by having
    # our mock open_url return a generic valid value, True.
    mock_open_url.return_value = True

    # Substitute in our fake results to be returned by get_results().
    # We need to have semi-valid looking URLs to pass internal checks,
    # but can provide trivially fake data.
    fake_results = {}
    fake_data_url = "https://fake.data"
    timestamp = time.strptime('20030201_040506', "%Y%m%d_%H%M%S")
    for release in releases:
        k = f"{URL_AUTOPKGTEST}/results/autopkgtest-{release}-{owner_name}-{ppa_name}/"
        fake_results[k] = []
        for arch in architectures:
            # We don't care about the triggers for the result, just the
            # number of results and their basic identity, so replace the
            # get_triggers() call to avoid it making any remote calls.
            result = Result(url=fake_data_url, time=timestamp, series=release, arch=arch, source=None)
            result.get_triggers = lambda: "x"
            fake_results[k].append(result)

    def fake_get_results(response, base_url, arches, sources):
        return fake_results[base_url]
    mock_get_results.side_effect = fake_get_results

    ppa = Ppa(ppa_name, owner_name)
    results = ppa.get_autopkgtest_results(releases, architectures)

    assert len(results) == expected_num_results
    for trigger_set in results:
        assert isinstance(trigger_set, dict)
        assert all(len(triggers) == expected_num_triggers for triggers in trigger_set.values())


@pytest.mark.parametrize('address, expected', [
    # Successful cases
    ('bb', (None, 'bb')),
    ('123', (None, '123')),
    ('a/123', ('a', '123')),
    ('ppa:A/bb', ('a', 'bb')),
    ('ppa:a/bb', ('a', 'bb')),
    ('ppa:ç/bb', ('ç', 'bb')),
    ('https://launchpad.net/~a/+archive/ubuntu/bb', ('a', 'bb')),
    ('https://launchpad.net/~a/+archive/ubuntu/bb/', ('a', 'bb')),
    ('https://launchpad.net/~a/+archive/ubuntu/bb////', ('a', 'bb')),
    ('https://launchpad.net/~a/+archive/ubuntu/bb/+packages', ('a', 'bb')),
    ('https://launchpad.net/~a/+archive/ubuntu/bb/+x', ('a', 'bb')),

    # Expected failure cases
    ('ppa:', (None, None)),
    (None, (None, None)),
    ('', (None, None)),
    ('/', (None, None)),
    (':/', (None, None)),
    ('////', (None, None)),
    ('ppa:/', (None, None)),
    ('ppa:a/', (None, None)),
    ('ppa:/bb', (None, None)),
    ('ppa:a/bç', (None, None)),
    ('ppa/a/bb', (None, None)),
    ('ppa:a/bb/c', (None, None)),
    ('ppa:a/bB', (None, None)),
    ('http://launchpad.net/~a/+archive/ubuntu/bb', (None, None)),
    ('https://example.com/~a/+archive/ubuntu/bb', (None, None)),
    ('https://launchpad.net/~a/+archive/nobuntu/bb', (None, None)),
    ('https://launchpad.net/~a/+archive/ubuntu/bb/x', (None, None)),
    ('https://launchpad.net/~a/+archive/ubuntu/bb/+', (None, None)),
])
def test_ppa_address_split(address, expected):
    """Check ppa address input strings can be parsed properly."""
    result = ppa_address_split(address)
    assert result == expected


def test_default_pocket():
    lp = LpServiceMock()
    ppa = PpaGroup(name='me', service=lp).create('foo')
    assert ppa.pocket == 'updates'


def test_set_pocket():
    lp = LpServiceMock()
    ppa = PpaGroup(name='me', service=lp).create('foo')
    ppa.set_pocket('proposed')
    assert ppa.pocket == 'proposed'


def test_set_pocket_error():
    lp = LpServiceMock()
    ppa = PpaGroup(name='me', service=lp).create('foo')
    with pytest.raises(ValueError):
        ppa.set_pocket('invalid-pocket')


def test_get_ppa():
    ppa = get_ppa(None, {'owner_name': 'a', 'ppa_name': 'bb'})
    assert type(ppa) is Ppa
    assert ppa.owner_name == 'a'
    assert ppa.ppa_name == 'bb'



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2022 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the Job class as an interface to autopkgtest runs."""

import os
import sys

import json
import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.job import Job, get_running, get_waiting
from tests.helpers import RequestResponseMock


def test_object():
    """Checks that Job objects can be instantiated."""
    job = Job(0, '', '', '', '')
    assert job


def test_init():
    """Checks the initialization of a Job object."""
    job = Job(0, 'a', 'b', 'c', 'd')
    assert job.number == 0
    assert job.submit_time == 'a'
    assert job.source_package == 'b'
    assert job.series == 'c'
    assert job.arch == 'd'


def test_repr():
    """Checks Job object machine-parsable representation."""
    job = Job(0, 'a', 'b', 'c', 'd')
    assert repr(job) == "Job(source_package='b', series='c', arch='d')"


def test_str():
    """Checks Job object textual presentation."""
    job = Job(0, 'a', 'b', 'c', 'd')
    assert f"{job}" == 'b c (d)'


def test_to_dict():
    """Checks Job object structural representation."""
    job = Job(0, 'a', 'b', 'c', 'd')
    expected_keys = [
        'number', 'submit_time', 'source_package_name',
        'series', 'arch', 'triggers', 'ppas'
    ]
    expected_types = [str, int, list]

    d = job.to_dict()
    assert isinstance(d, dict), f"type of d is {type(d)} not dict"

    # Verify expected keys are present
    assert sorted(d.keys()) == sorted(expected_keys)

    # Verify values are within set of expected types
    for k, v in d.items():
        assert type(v) in expected_types, f"'{k}={v}' is unexpected type {type(v)}"

    # Verify full dict can be written as JSON
    try:
        assert json.dumps(d)
    except UnicodeDecodeError as e:
        assert False, f"Wrong UTF codec detected: {e}"
    except json.JSONDecodeError as e:
        assert False, f"JSON decoding error: {e.msg}, {e.doc}, {e.pos}"


def test_triggers():
    """Checks Job object's triggers."""
    job = Job(0, '', '', '', '', triggers=['a/1', 'b/2'])
    assert job.triggers == ['a/1', 'b/2']


def test_ppas():
    """Checks Job object textual presentation."""
    job = Job(0, '', '', '', '', ppas=['ppa:a/b', 'ppa:c/d'])
    assert job.ppas == ['ppa:a/b', 'ppa:c/d']


@pytest.mark.parametrize('triggers, endswith', [
    (['t/1'], '?release=c&arch=d&package=b&trigger=t%2F1'),
    (['t/2.3'], '?release=c&arch=d&package=b&trigger=t%2F2.3'),
    (['t/2.3-1'], '?release=c&arch=d&package=b&trigger=t%2F2.3-1'),
    (['t/2.3-1ubuntu2'], '?release=c&arch=d&package=b&trigger=t%2F2.3-1ubuntu2'),
    (['t/2.3-1ubuntu2~ppa1'], '?release=c&arch=d&package=b&trigger=t%2F2.3-1ubuntu2~ppa1'),
    (['t/2.3-1ubuntu2+ppa2'], '?release=c&arch=d&package=b&trigger=t%2F2.3-1ubuntu2%2Bppa2'),
    (['t/1', 'u/2'], '?release=c&arch=d&package=b&trigger=t%2F1&trigger=u%2F2'),
])
def test_request_url(triggers, endswith):
    """Checks Job object can create valid URLs for starting the test runs."""
    jobinfo = {
        'triggers': triggers,
        'ppas': ['ppa:a/b', 'ppa:c/d']
    }
    job = Job(0, 'a', 'b', 'c', 'd', jobinfo['triggers'], jobinfo['ppas'])
    assert job.request_url.startswith("https://autopkgtest.ubuntu.com/request.cgi")
    assert job.request_url.endswith(endswith)


@pytest.mark.parametrize('json_text, params, expected', [
    (
        # No Jobs returned if nothing is running in autopkgtest
        '',
        {'releases': [], 'sources': None},
        {'repr': '', 'triggers': [], 'ppas': []}
    ),
    (
        # No Jobs returned if nothing is running for the specified PPA
        (
            '{"x": {"x-job-id": {"focal": { "x": ['
            '{"submit-time": "2022-08-19 20:59:01", '
            '"triggers": ["x/1.2.3"], '
            '"ppas": ["ppa:x/x"]}, '
            '1234, '
            '"Log Output Here"'
            '] } } } }'
        ),
        {'releases': [], 'ppa': 'ppa:me/myppa', 'sources': None},
        {'repr': '', 'triggers': [], 'ppas': []}
    ),
    (
        # No Jobs returned if nothing running for the PPA matches the specified release
        (
            '{"x": {"x-job-id": {"focal": { "x": ['
            '{"submit-time": "2022-08-19 20:59:01", '
            '"triggers": ["x/1.2.3"], '
            '"ppas": ["ppa:me/myppa"]}, '
            '1234, '
            '"Log Output Here"'
            '] } } } }'
        ),
        {'releases': ['jammy'], 'ppa': 'ppa:me/myppa', 'sources': None},
        {'repr': '', 'triggers': [], 'ppas': []}
    ),
    (
        # No Jobs returned if nothing running for the PPA matches the specified package
        (
            '{"x": {"x-job-id": {"focal": { "arm64": ['
            '{"submit-time": "2022-08-19 20:59:01", '
            '"triggers": ["x/1.2.3"], '
            '"ppas": ["ppa:me/myppa"]}, '
            '1234, '
            '"Log Output Here"'
            '] } } } }'
        ),
        {'releases': None, 'ppa': 'ppa:me/myppa', 'sources': ['mypackage']},
        {'repr': '', 'triggers': [], 'ppas': []}
    ),
    (
        # Correct Job should be returned for matching criteria
        (
            '{"mypackage": {"x-job-id": {"focal": { "arm64": ['
            '{"submit-time": "2022-08-19 20:59:01", '
            '"triggers": ["x/1.2.3"], '
            '"ppas": ["ppa:me/myppa"]}, '
            '1234, '
            '"Log Output Here"'
            '] } } } }'
        ),
        {'releases': ['focal'], 'ppa': 'ppa:me/myppa', 'sources': None},
        {
            'repr': "Job(source_package='mypackage', series='focal', arch='arm64')",
            'triggers': ["x/1.2.3"],
            'ppas': ["ppa:me/myppa"]
        }
    ),
])
def test_get_running(json_text, params, expected):
    """Checks result of the get_running() command."""
    fake_response = RequestResponseMock(json_text)

    try:
        job = next(get_running(fake_response, **params))
        assert repr(job) == expected['repr']
        assert job.triggers == expected['triggers']
        assert job.ppas == expected['ppas']
    except StopIteration:
        assert expected['repr'] == ''
        assert expected['triggers'] == []
        assert expected['ppas'] == []


@pytest.mark.parametrize('json_text, params, expected', [
    (
        # No Jobs returned if nothing is running in autopkgtest
        "",
        {'releases': [], 'sources': None},
        {'repr': '', 'source_package': [], 'triggers': [], 'ppas': []}
    ),
    (
        # No Jobs returned if nothing is running for the specified PPA
        (
            '{ "ubuntu": { "focal": { "amd64": ['
            ' "a\\n{\\"requester\\":   \\"you\\",'
            '      \\"submit-time\\": \\"2022-08-19 07:37:56\\",'
            '      \\"triggers\\":    [ \\"a/1.2-3\\", \\"b/1-1\\" ] }",'
            ' "b\\n{\\"requester\\":   \\"you\\",'
            '      \\"submit-time\\": \\"2022-08-19 07:37:57\\",'
            '      \\"ppas\\":        [ \\"ppa:x/x\\" ],'
            '      \\"triggers\\":    [ \\"c/3.2-1\\", \\"d/2-2\\" ] }"'
            '] } } }'
        ),
        {'releases': None, 'ppa': 'ppa:me/myppa', 'sources': None},
        {'repr': '', 'source_package': None, 'triggers': [], 'ppas': []}
    ),
    (
        # Correct Job should be returned for matching criteria
        (
            '{ "ubuntu": { "focal": { "amd64": ['
            ' "a\\n{\\"requester\\":   \\"you\\",'
            '      \\"submit-time\\": \\"2022-08-19 07:37:56\\",'
            '      \\"triggers\\":    [ \\"a/1.2-3\\", \\"b/1-1\\" ] }",'
            ' "b\\n{\\"requester\\":   \\"you\\",'
            '      \\"submit-time\\": \\"2022-08-19 07:37:57\\",'
            '      \\"ppas\\":        [ \\"ppa:me/myppa\\" ],'
            '      \\"triggers\\":    [ \\"c/3.2-1\\", \\"d/2-2\\" ] }"'
            '] } } }'
        ),
        {'releases': ['focal'], 'ppa': 'ppa:me/myppa', 'sources': None},
        {
            'repr': "Job(source_package='b', series='focal', arch='amd64')",
            'source_package': 'b',
            'triggers': ['c/3.2-1', 'd/2-2'],
            'ppas': ['ppa:me/myppa']
        }
    ),
])
def test_get_waiting(json_text, params, expected):
    """Checks result of the get_waiting() command."""
    # TODO: I think ppas need to be in "ppa" instead of under "ubuntu" but need to doublecheck.
    fake_response = RequestResponseMock(json_text)
    try:
        job = next(get_waiting(fake_response, **params))
        assert job

        assert repr(job) == expected['repr']
        assert job.source_package == expected['source_package']
        assert job.triggers == expected['triggers']
        assert job.ppas == expected['ppas']
    except StopIteration:
        assert expected['repr'] == ''
        assert expected['triggers'] == []
        assert expected['ppas'] == []



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2022 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests Results class as representing an Autopkgtest test log and status."""

import os
import sys
import time

import gzip
import json
import pytest
from typing import List, Dict

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))
DATA_DIR = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "data"))

from ppa.result import Result, get_results, show_results
from ppa.subtest import Subtest
from ppa.io import open_url


def test_object():
    """Checks that Result objects can be instantiated."""
    timestamp = time.strptime('20030201_040506', "%Y%m%d_%H%M%S")
    result = Result('url', timestamp, 'ser', 'arch', 'src')
    assert result
    assert result.url == 'url'
    assert result.time == timestamp
    assert result.series == 'ser'
    assert result.arch == 'arch'
    assert result.source == 'src'
    assert not result.error_message


@pytest.mark.parametrize('pub_dict, expected_repr', [
    (
        {'url': 'url', 'time': None, 'series': 'ser', 'arch': 'arch', 'source': 'src'},
        "Result(url='url', time=None, series='ser', arch='arch', source='src')",
    )
])
def test_repr(pub_dict, expected_repr):
    """Checks Result object representation."""
    result = Result(**pub_dict)
    assert repr(result) == expected_repr


@pytest.mark.parametrize('pub_dict, expected_str', [
    (
        {
            'url': 'url',
            'time': time.strptime('20030201_040506', "%Y%m%d_%H%M%S"),
            'series': 'ser',
            'arch': 'arch',
            'source': 'src',
        },
        'src on ser for arch    @ 01.02.03 04:05:06',
    )
])
def test_str(pub_dict, expected_str):
    """Checks Result object textual presentation."""
    result = Result(**pub_dict)
    assert f"{result}" == expected_str


@pytest.mark.parametrize('timestamp, subtests, show_urls, expected', [
    (
        # With show_urls enabled, output should show a text URL.
        '20030201_040506',
        [],
        True,
        """    + ✅ source on series for arch    @ 01.02.03 04:05:06
      • Log: file://<DATA_DIR>/x
"""
    ),
    (
        # Without show_urls, output should create console-clickable URLs.
        '20030201_040506',
        [],
        False,
        """    + ✅ source on series for arch    @ 01.02.03 04:05:06  \x1b]8;;file://<DATA_DIR>/x\x1b\\Log️ 🗒️ \x1b]8;;\x1b\\
"""
    ),
    (
        # When no Subtests fail, no detailed results should be shown.
        '20030201_040506',
        [Subtest('x PASS'), Subtest('y PASS'), Subtest('z SKIP')],
        True,
        """    + ✅ source on series for arch    @ 01.02.03 04:05:06
      • Log: file://<DATA_DIR>/x
"""
    ),
    (
        # When at least one Subtest fails, all subtest results should be shown.
        '20030201_040506',
        [Subtest('x PASS'), Subtest('y FAIL'), Subtest('f FLAKY'), Subtest('z SKIP')],
        True,
        """    + ❌ source on series for arch    @ 01.02.03 04:05:06
      • Log: file://<DATA_DIR>/x
      • Status: FAIL
      • x                         PASS   🟩
      • y                         FAIL   🟥
      • f                         FLAKY  🟫
      • z                         SKIP   🟧
"""
    ),
])
def test_to_bullet_tree(timestamp, subtests, show_urls, expected):
    """Checks representation of Result as a bullet-tree text list."""
    tm = time.strptime(timestamp, "%Y%m%d_%H%M%S")
    result = Result(f"file://{DATA_DIR}/x", tm, 'series', 'arch', 'source')

    # Substitute in our fake subtest data in place of Result's get_subtests() routine
    result.get_subtests = lambda: subtests

    out = result.to_bullet_tree(show_urls)
    assert out == expected.replace("<DATA_DIR>", DATA_DIR)


def test_timestamp():
    """Checks Result object formats the result's time correctly."""
    timestamp = time.strptime('20030201_040506', "%Y%m%d_%H%M%S")
    result = Result('url', timestamp, 'b', 'c', 'd')
    assert f"{result.timestamp}" == '01.02.03 04:05:06'


def test_log(tmp_path):
    """Checks that the log content of a Result is available."""
    f = tmp_path / "result.log.gz"
    compressed_text = gzip.compress(bytes('abcde', 'utf-8'))
    f.write_bytes(compressed_text)

    result = Result(f"file://{f}", None, None, None, None)
    assert result.log == "abcde"


@pytest.mark.parametrize('filename, expected_triggers', [
    ('results-rabbitmq-server-armhf.log.gz', ['rabbitmq-server/3.9.27-0ubuntu0.1~jammy8']),
    ('results-six-s390x.log.gz', ['pygobject/3.42.2-2', 'six/1.16.0-4']),
    ('results-chrony-armhf.log.gz', ['dpkg/1.22.6ubuntu5'])
])
def test_triggers(filename, expected_triggers):
    """Checks that autopkgtest triggers can be extracted from test result logs."""
    result = Result(f"file://{DATA_DIR}/{filename}", None, None, None, None)
    assert result.triggers == expected_triggers


@pytest.mark.parametrize('triggers, name, arch, series, expected', [
    ([], None, 'arch', 'ser', []),
    (
        # Single trigger
        ['pkg/1'],
        'pkg', 'arch', 'ser',
        [
            (
                "Trigger(source_publication=SourcePublication(package='pkg', "
                "version='1', series='ser', archive=None, status=None), "
                "architecture='arch', test_package='pkg')"
            )
        ],
    ),
    pytest.param(
        # Multiple triggers
        ['pkg/1', 'x/2', 'y/3', 'z/4'],
        'pkg', 'arch', 'ser',
        [
            (
                "Trigger(source_publication=SourcePublication(package='pkg', "
                "'version='1', series='ser', archive=None, status=None), "
                "architecture='arch', test_package='pkg')"
            ),
            (
                "Trigger(source_publication=SourcePublication(package='x', "
                "'version='2', series='ser', archive=None, status=None), "
                "architecture='arch', test_package='pkg')"
            ),
            (
                "Trigger(source_publication=SourcePublication(package='y', "
                "'version='3', series='ser', archive=None, status=None), "
                "architecture='arch', test_package='pkg')"
            ),
            (
                "Trigger(source_publication=SourcePublication(package='z', "
                "'version='4', series='ser', archive=None, status=None), "
                "architecture='arch', test_package='pkg')"
            ),
        ],
        marks=pytest.mark.xfail(reason="Trigger and test packages reversed?"),
    ),
])
def test_get_triggers(monkeypatch, triggers, name, arch, series, expected):
    """Checks retrieval of Trigger objects from autopkgtest results."""
    result = Result('url', time=None, arch=arch, series=series, source=None)
    monkeypatch.setattr(Result, "triggers", triggers)

    triggers = result.get_triggers(name)
    assert [repr(t) for t in triggers] == expected


@pytest.mark.parametrize('log_text, subtest_name, expected', [
    ('', None, {'testbed': 'BAD'}),
    (
        (
            "x: @@@@@@@@@@@@@@@@@@@@ summary\n"
            "test-a          PASS\n"
            "test-b          FAIL ignorable-note\n"
            "test-c          FLAKY some-detail\n"
            "test-d          NoTaVaLiDsTaTuS\n"
        ),
        None,
        {'test-a': 'PASS', 'test-b': 'FAIL', 'test-c': 'FLAKY'}
    ),
    (
        (
            "autopkgtest [21:13:56]: starting date: 2022-11-18\n"
            "The following packages have unmet dependencies:\n"
            " builddeps:.../12-autopkgtest-satdep.dsc:i386 : Depends: gcc:i386 but it is not installable\n"
            "E: Unable to correct problems, you have held broken packages.\n"
            "chroot               FAIL badpkg\n"
            "blame: apache2\n"
            "badpkg: Test dependencies are unsatisfiable. A common reason is ...\n"
            "autopkgtest [21:48:03]: @@@@@@@@@@@@@@@@@@@@ summary\n"
            "run-test-suite       FAIL badpkg\n"
            "blame: apache2\n"
            "badpkg: Test dependencies are unsatisfiable. A common reason is...\n"
            "duplicate-module-load PASS\n"
            "default-mods         PASS\n"
            "run-htcacheclean     PASS\n"
            "ssl-passphrase       PASS\n"
            "check-http2          PASS\n"
            "run-chroot           FAIL badpkg\n"
            "blame: apache2\n"
        ),
        'run-',
        {
            'run-test-suite': 'FAIL',
            'run-htcacheclean': 'PASS',
            'run-chroot': 'FAIL',
        }
    ),
    (
        (
            "3657s rm: cannot remove '.../mountpoint': Device or resource busy\n"
            "3661s autopkgtest [03:41:43]: test minimized: -----------------------]\n"
            "3663s autopkgtest [03:41:45]: test minimized:  - - - - - - - - - - results - - - - - - - - - -\n"
            "3663s minimized            FAIL non-zero exit status 1\n"
            "3663s autopkgtest [03:41:45]: test minimized:  - - - - - - - - - - stderr - - - - - - - - - -\n"
            "3663s rm: cannot remove '.../mountpoint': Device or resource busy\n"
            "3664s autopkgtest [03:41:46]: @@@@@@@@@@@@@@@@@@@@ summary\n"
            "3664s default-bootstraps   FAIL non-zero exit status 1\n"
            "3664s minimized            FAIL non-zero exit status 1'\n"
        ),
        None,
        {
            'default-bootstraps': 'FAIL',
            'minimized': 'FAIL'
        }
    ),
])
def test_get_subtests(tmp_path, log_text: str, subtest_name: str, expected: Dict[str, str]):
    """Checks retrieval of Subtest objects from autopkgtest results.

    This test exercises the parser that extracts subtest information out
    of autopkgtest logs of various formats.  It also verifies the
    parameter to get_subtests() is handled correctly.

    :param fixture tmp_path: Temp dir.
    :param str log_text: Text to write into the log file.
    :param str subtest_name: Only retrieve subtests starting with this text.
    :param Dict[str] expected: Dictionary of subtest names to pass/fail status.
    """
    f = tmp_path / "result.log.gz"
    compressed_text = gzip.compress(bytes(log_text, 'utf-8'))
    f.write_bytes(compressed_text)

    result = Result(f"file://{f}", None, None, None, None)
    subtests = result.get_subtests(subtest_name)
    assert {s.desc: s.status for s in subtests} == expected


@pytest.mark.parametrize('subtest_states, error_message, expected', [
    ([], None, 'PASS'),
    (['PASS'], None, 'PASS'),
    (['FAIL'], None, 'FAIL'),
    (['SKIP'], None, 'PASS'),
    (['FLAKY'], None, 'PASS'),
    (['PASS', 'FAIL'], None, 'FAIL'),
    (['FAIL', 'PASS'], None, 'FAIL'),
    (['PASS', 'PASS', 'PASS'], None, 'PASS'),
    (['PASS', 'FLAKY'], None, 'PASS'),
    (['PASS', 'SKIP'], None, 'PASS'),
    (['PASS'], 'x', 'BAD'),
    (['FAIL'], 'x', 'BAD'),
])
def test_status(monkeypatch, subtest_states, error_message, expected):
    """Checks retrieval of status from autopkgtest results."""
    result = Result("file://tmp/x", None, None, None, None)
    result.error_message = error_message

    # Add subtests with given states
    subtests = [Subtest(f"... {state}...") for state in subtest_states]
    monkeypatch.setattr(Result, "get_subtests", lambda x: subtests)

    assert result.status == expected


@pytest.mark.parametrize('status, expected', [
    ('PASS', "✅"),
    ('FAIL', "❌"),
    ('BAD', "⛔"),
])
def test_status_icon(monkeypatch, status, expected):
    """Checks generation of correct icon based on autopkgtest results."""
    result = Result("file://tmp/x", None, None, None, None)
    monkeypatch.setattr(Result, "status", status)

    assert result.status_icon == expected


@pytest.mark.parametrize('status, expected_exception', [
    (None, KeyError),
    ('x', KeyError),
])
def test_status_icon_error(monkeypatch, status, expected_exception):
    """Checks generation of correct icon based on autopkgtest results."""
    result = Result("file://tmp/x", None, None, None, None)
    monkeypatch.setattr(Result, "status", status)

    with pytest.raises(expected_exception):
        print(result.status_icon)


@pytest.mark.parametrize('log_text, series, arch, source, expected_dict', [
    (
        # Empty/invalid log should return empty triggers
        'x', 'x', 'x', 'x', {'triggers': []}
    ),

    (
        # Empty/invalid log counts as an overall test state BAD.
        'x', 'x', 'x', 'x',
        {
            'status': 'BAD',
            'status_icon': '⛔',
            'subtests': [
                {
                    'desc': 'testbed',
                    'line': 'testbed setup failure  BAD',
                    'status': 'BAD',
                    'status_icon': '⛔'
                }
            ]
        }
    ),

    (
        # Init parameters are registered in the class as provided.
        'l', 's', 'a', 'pkg', {'log': 'l', 'series': 's', 'arch': 'a', 'source': 'pkg'}
    ),

    (
        # Log with valid syntax for a trigger should create a Trigger dict.
        '--env=ADT_TEST_TRIGGERS=t/1 -- \n: @@@@@@@@@@@@@@@@@@@@', 's', 'a', 'pkg',
        {
            'triggers': [
                {
                    'source_publication': {
                        'archive': None,
                        'package': 't',
                        'version': '1',
                        'status': None,
                        'series': 's',
                    },
                    'architecture': 'a',
                    'test_package': 't',
                },
            ]
        }
    ),

    (
        # Log with valid syntax for a subtest should create a Subtest dict.
        ': @@@@@@@@@@@@@@@@@@@@ summary\n999s tst FAIL', 's', 'a', 'pkg',
        {
            'subtests': [
                {
                    'desc': 'tst',
                    'line': 'tst FAIL',
                    'status': 'FAIL',
                    'status_icon': '🟥'
                },
            ],
            'status': 'FAIL',
            'status_icon': '❌'
        }
    )
])
def test_to_dict(tmp_path, log_text, series, arch, source, expected_dict):
    """Checks Result object structural representation."""
    f = tmp_path / "result.log.gz"
    f.write_bytes(gzip.compress(bytes(log_text, 'utf-8')))
    timestamp = time.strptime('20030201_040506', "%Y%m%d_%H%M%S")
    result = Result(f"file://{f}", timestamp, series, arch, source)
    expected_keys = [
        'url', 'timestamp', 'series', 'arch', 'source',
        'error_message', 'log', 'triggers', 'subtests', 'status',
        'status_icon'
    ]
    expected_types = [str, type(None), list]

    d = result.to_dict()
    assert isinstance(d, dict), f"type of d is {type(d)} not dict"

    # Verify expected keys are present
    assert sorted(d.keys()) == sorted(expected_keys)

    # Verify values are within set of expected types
    for k, v in d.items():
        assert type(v) in expected_types, f"'{k}={v}' is unexpected type {type(v)}"

    # Verify values match what we expect
    for k, v in expected_dict.items():
        assert v == d.get(k)

    # Verify full dict can be written as JSON
    try:
        assert json.dumps(d)
    except UnicodeDecodeError as e:
        assert False, f"Wrong UTF codec detected: {e}"
    except json.JSONDecodeError as e:
        assert False, f"JSON decoding error: {e.msg}, {e.doc}, {e.pos}"


@pytest.mark.parametrize('results_text, arches, sources, expected', [
    (
        # Specifying no args should return all logs
        "noble/amd64/e/exim4/20240104_193939_e912d@/log.gz\n",
        None,
        None,
        ["/results/noble/amd64/e/exim4/20240104_193939_e912d@/log.gz'"],
    ),
    (
        # Specifying empty args should return all logs
        "noble/amd64/e/exim4/20240104_193939_e912d@/log.gz\n",
        [],
        [],
        ["/results/noble/amd64/e/exim4/20240104_193939_e912d@/log.gz'"],
    ),
    (
        # Specifying a missing architecture should return no results
        "noble/amd64/e/exim4/20240104_193939_e912d@/log.gz",
        ['a'],
        None,
        []
    ),
    (
        # Specifying an unrelated package should return no results
        "noble/amd64/e/exim4/20240104_193939_e912d@/log.gz\n",
        None,
        ['pkg'],
        []
    ),
    (
        # Specifying an arch and package present should return
        # corresponding result
        'noble/amd64/e/exim4/20240104_193939_e912d@/log.gz',
        ['amd64'],
        ['exim4'],
        ["/results/noble/amd64/e/exim4/20240104_193939_e912d@/log.gz'"],
    ),
    (
        # Specified architectures should provide results for those
        # architectures and no others.
        (
            "noble/amd64/e/exim4/20240104_193939_e912d@/log.gz\n"
            "noble/armhf/e/exim4/20240104_193939_e912d@/log.gz\n"
            "noble/arm64/e/exim4/20240104_193939_e912d@/log.gz\n"
            "noble/i386/e/exim4/20240104_193939_e912d@/log.gz\n"
        ),
        ['amd64', 'armhf', 'i386'],
        ['exim4'],
        [
            "/results/noble/amd64/e/exim4/20240104_193939_e912d@/log.gz'",
            "/results/noble/armhf/e/exim4/20240104_193939_e912d@/log.gz'",
            "/results/noble/i386/e/exim4/20240104_193939_e912d@/log.gz'",
        ],
    ),
    (
        # Specifying mix of present and missing architectures should
        # correctly return proper subset of matches.
        (
            "noble/amd64/e/exim4/20240104_193939_e912d@/log.gz\n"
            "noble/armhf/e/exim4/20240104_193939_e912d@/log.gz\n"
            "noble/arm64/e/exim4/20240104_193939_e912d@/log.gz\n"
            "noble/i386/e/exim4/20240104_193939_e912d@/log.gz\n"
        ),
        ['amd64', 'x', 'y', 'z'],
        ['exim4'],
        [
            "/results/noble/amd64/e/exim4/20240104_193939_e912d@/log.gz'",
        ],
    ),
    (
        # Specifying a mix of present and missing packages should
        # correctly return proper subset of matches.
        (
            "noble/amd64/x/xxx/20240104_193939_e912d@/log.gz\n"
            "noble/amd64/y/yyy/20240104_193939_e912d@/log.gz\n"
        ),
        ['amd64'],
        ['xxx', 'yyy', 'zzz', 'non-existing'],
        [
            "/results/noble/amd64/x/xxx/20240104_193939_e912d@/log.gz'",
            "/results/noble/amd64/y/yyy/20240104_193939_e912d@/log.gz'",
        ],
    ),
])
def test_get_results(tmp_path,
                     results_text: str,
                     arches: 'List[str]|None',
                     sources: 'List[str]|None',
                     expected: List[str]):
    """Checks that expected results can be found from autopkgtest logs."""
    f = tmp_path / "results"
    f.write_text(results_text)
    url = f"file://{f}"
    response = open_url(url)
    assert response

    results = list(get_results(response, url, arches, sources))
    print("results: ", results)

    # Verify expected number, type, and representation of returned results
    assert len(results) == len(expected)
    assert all([isinstance(result, Result) for result in results])
    for r in results:
        # Verify the repr starts with the expected class name
        assert repr(r).startswith("Result(url='")

        # Verify that at least one of the expected strings is present
        # somewhere in the repr output string.
        assert True in [e in repr(r) for e in expected], \
            f"None of {expected} found in {repr(r)}"


@pytest.mark.parametrize('data, show_urls, expected_num_lines, expected_in_stdout', [
    (
        # When there are no results, indicate '(none)'
        {}, True, 1, "* Results: (none)\n"
    ),
    (
        {'trigger-1': ('a', 'b', 'c', 'x')}, True, 6, "- trigger-1\n    + ⛔ c on a for b"
    ),
    ({'t1': ('a', 'b', 'c', 'x'), 't2': ('a', 'b', 'c', 'x')}, True, 11, "- t2"),
])
def test_show_results(capfd, tmp_path,
                      data, show_urls, expected_num_lines, expected_in_stdout):
    """Checks that results output includes the expected text."""
    results = []
    for trigger, (series, arch, source, item) in data.items():
        trigger_sets = {}
        trigger_sets.setdefault(trigger, [])
        f = tmp_path / f"{trigger}-result.log.gz"
        compressed_text = gzip.compress(bytes(item, 'utf-8'))
        f.write_bytes(compressed_text)
        timestamp = time.strptime('20030201_040506', "%Y%m%d_%H%M%S")
        result = Result(f"file://{f}", timestamp, series, arch, source)
        trigger_sets[trigger].append(result)
        results.append(trigger_sets)

    show_results(results, show_urls)

    out, err = capfd.readouterr()
    print(out)
    assert out.count('\n') == expected_num_lines
    assert expected_in_stdout in out



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2023 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""Interprets and analyzes an Ubuntu Apt record for a source package."""

from functools import lru_cache
from typing import Dict

from .dict import unpack_to_dict


class SourcePackage:
    """The Apt record for the packaged source of a software codebase.

    This object is intended to be instantiated from a raw Apt source
    package record such as extracted from a Sources.xz file by
    apt_pkg.TagFile().  The SourcePackage class members handle the
    parsing and interpretation of the data values of the record such as
    build dependencies.

    Since objects of this class are constructed as thin wrappers on the
    Apt record, the available member properties for the object will vary
    depending on the data elements present in the record.  For example,
    if there isn't a "Testsuite:" field, then this object will not have a
    'self.testsuite' property.

    For member properties needed by users of this class, it's
    recommended to guarantee their presence either by requiring them
    from the pkg_dict in the __init__() property, or add a property to
    the class that checks and substitutes a suitable default if missing.
    """
    # pylint: disable=no-member

    def __init__(self, pkg_dict: dict):
        """Initialize a new SourcePackage object for a given package.

        :param dict[str, str] pkg_dict: Data collection loaded from am Apt record.
        """
        # Auto-create member parameters from the Apt record's data structure
        for k, val in dict(pkg_dict).items():
            setattr(self, k.lower().replace('-', '_'), val)

        for field in ['package', 'version', 'binary']:
            if getattr(self, field, None) is None:
                raise ValueError(f'undefined {field} from Apt record for source package')

    def __repr__(self):
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'pkg_dict={vars(self)!r})')

    def __str__(self) -> str:
        """Return a human-readable textual description of the SourcePackage.

        :rtype: str
        :returns: Human-readable string.
        """
        return f"{self.package} ({self.version})"

    @property
    def name(self) -> str:
        """The name of the source package as recorded in the Apt database.

        :rtype: str
        :returns: Package name.
        """
        return self.package

    @property
    @lru_cache
    def provides_binaries(self) -> Dict[str, str]:
        """The binary package names provided by this source package.

        For consistency, the results are provided as a dict structure,
        however since the Binary field is a simple comma-separated list,
        the dict will only have keys, with no defined values.

        :rtype: dict[str, str]
        :returns: Collection of binary packages this source provides.
        """
        bins = unpack_to_dict(self.binary)
        if not bins:
            raise RuntimeError('Could not parse packages from Binaries line of Apt record.')
        return bins

    @property
    @lru_cache
    def build_dependencies(self) -> Dict[str, str]:
        """The binary packages that must be available to build this source package.

        :rtype: dict[str, BinaryPackages]
        :returns: Collection of package names to version specification strings.
        """
        if getattr(self, 'build_depends', None) is None:
            # Missing BuildDepends is uncommon, but can occur for packages
            # consisting just of config files, for example.
            return {}
        deps = unpack_to_dict(self.build_depends, key_sep=' ')
        if not deps:
            raise RuntimeError('Could not parse packages from Build-Depends line of Apt record.')
        return deps


if __name__ == "__main__":
    # pylint: disable=line-too-long, invalid-name

    import os
    import apt_pkg

    from pprint import PrettyPrinter
    pp = PrettyPrinter(indent=4)

    root_dir = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
    tests_data_dir = os.path.join(root_dir, 'tests', 'data')
    assert os.path.exists(tests_data_dir), f'cannot find {tests_data_dir}'

    print('####################################')
    print('## SourcePackage class smoke test ##')
    print('####################################')

    print()
    print('Source Packages')
    print('---------------')
    with apt_pkg.TagFile(f'{tests_data_dir}/source/Sources.xz') as pkgs:
        results = []
        for pkg in pkgs:
            source = SourcePackage(pkg)
            results.append(str(source))

        num_results = len(results)
        ellipses_shown = False
        for i, result in enumerate(results):
            if i < 10 or i >= num_results - 10:
                print(result)
            elif not ellipses_shown:
                print('...')
                ellipses_shown = True
    print()

    print()
    print('Details for "dovecot"')
    print('---------------------')
    package_data = {
        'Package': 'dovecot',
        'Format': '3.0 (quilt)',
        'Binary': 'dovecot-core, dovecot-dev, dovecot-imapd, dovecot-pop3d, dovecot-lmtpd, dovecot-managesieved, dovecot-pgsql, dovecot-mysql, dovecot-sqlite, dovecot-ldap, dovecot-gssapi, dovecot-sieve, dovecot-solr, dovecot-lucene, dovecot-submissiond, dovecot-auth-lua',  # noqa: E501
        'Architecture': 'any',
        'Version': '1:2.3.19.1+dfsg1-2ubuntu4',
        'Priority': 'optional',
        'Section': 'mail',
        'Maintainer': 'Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>',
        'Original-Maintainer': 'Dovecot Maintainers <dovecot@packages.debian.org>',
        'Uploaders': 'Jaldhar H. Vyas <jaldhar@debian.org>, Jelmer Vernooij <jelmer@debian.org>, Apollon Oikonomopoulos <apoikos@debian.org>, Noah Meyerhans <noahm@debian.org>',  # noqa: E501
        'Standards-Version': '4.6.1',
        'Build-Depends': 'debhelper-compat (= 13), default-libmysqlclient-dev, krb5-multidev, libapparmor-dev [linux-any], libbz2-dev, libcap-dev [linux-any], libclucene-dev, libdb-dev, libexpat-dev, libexttextcat-dev, libicu-dev, libldap2-dev, liblua5.3-dev, liblz4-dev, liblzma-dev, libpam0g-dev, libpq-dev, libsasl2-dev, libsodium-dev, libsqlite3-dev, libssl-dev, libstemmer-dev, libsystemd-dev [linux-any], libunwind-dev [amd64 arm64 armel armhf hppa i386 ia64 mips mips64 mips64el mipsel powerpc powerpcspe ppc64 ppc64el sh4], libwrap0-dev, libzstd-dev, lsb-release, pkg-config, zlib1g-dev',  # noqa: E501
        'Testsuite': 'autopkgtest',
        'Homepage': 'https://dovecot.org/',
        'Vcs-Browser': 'https://salsa.debian.org/debian/dovecot',
        'Vcs-Git': 'https://salsa.debian.org/debian/dovecot.git',
        'Directory': 'pool/main/d/dovecot',
        'Package-List': """
         dovecot-auth-lua deb mail optional arch=any
         dovecot-core deb mail optional arch=any
         dovecot-dev deb mail optional arch=any
         dovecot-gssapi deb mail optional arch=any
         dovecot-imapd deb mail optional arch=any
         dovecot-ldap deb mail optional arch=any
         dovecot-lmtpd deb mail optional arch=any
         dovecot-lucene deb mail optional arch=any
         dovecot-managesieved deb mail optional arch=any
         dovecot-mysql deb mail optional arch=any
         dovecot-pgsql deb mail optional arch=any
         dovecot-pop3d deb mail optional arch=any
         dovecot-sieve deb mail optional arch=any
         dovecot-solr deb mail optional arch=any
         dovecot-sqlite deb mail optional arch=any
         dovecot-submissiond deb mail optional arch=any
         """,  # noqa: E501
        'Files': """
         146ac1a3b2a90d96b7fe8c458a561ae2 3977 dovecot_2.3.19.1+dfsg1-2ubuntu4.dsc
         f8c84c45b05352d55c3dbd509389cc25 1636590 dovecot_2.3.19.1+dfsg1.orig-pigeonhole.tar.gz
         c334e8ef30546af8a668437f046f3f15 7790851 dovecot_2.3.19.1+dfsg1.orig.tar.gz
         c3e7d443a8f4854897d338a56f85c936 67620 dovecot_2.3.19.1+dfsg1-2ubuntu4.debian.tar.xz
         """,  # noqa: E501
        'Checksums-Sha1': """
         481053992c60990f765e0f671e6fc08e910a50b7 3977 dovecot_2.3.19.1+dfsg1-2ubuntu4.dsc
         cdf68b407f1237e92987c6353c9596f3458e2126 1636590 dovecot_2.3.19.1+dfsg1.orig-pigeonhole.tar.gz
         a35f87db78847ba469d9b0e3b72f15f8c5d1d9b0 7790851 dovecot_2.3.19.1+dfsg1.orig.tar.gz
         b50aac6b2aac52744e7ef459030bfefd15cedc71 67620 dovecot_2.3.19.1+dfsg1-2ubuntu4.debian.tar.xz
         """,  # noqa: E501
        'Checksums-Sha256': """
         560f603209443dc92f5941dc1a5b737e3a743defc45d56a1a48df3846deed0f8 3977 dovecot_2.3.19.1+dfsg1-2ubuntu4.dsc
         9bc08c0eeefd75452033e022936968ff0ddad037672effdffc4a7c8dd360b8e0 1636590 dovecot_2.3.19.1+dfsg1.orig-pigeonhole.tar.gz
         db5abcd87d7309659ea6b45b2cb6ee9c5f97486b2b719a5dd05a759e1f6a5c51 7790851 dovecot_2.3.19.1+dfsg1.orig.tar.gz
         7477feb66c8b8b4bab6fa4aadaab8675d308d575e5af662769e179a4fd6e289f 67620 dovecot_2.3.19.1+dfsg1-2ubuntu4.debian.tar.xz
         """,  # noqa: E501
        'Checksums-Sha512': """
         26ab2a3b3a29ee0734ffccbad826b68b8b290715b0dbebdc07e87bac66c5be07cb0bc303c8795fb9dab8c90c9802887ac7c2ddff4b4d789bbc61080a47379e1b 3977 dovecot_2.3.19.1+dfsg1-2ubuntu4.dsc
         a3d1ebab2896954d159af11df3de3993493ae304343fa41d7f408427f8bd19b3061a09017cfe2b62145b4aacaed87b1d9718418b3c73ae9ea700865844e5af39 1636590 dovecot_2.3.19.1+dfsg1.orig-pigeonhole.tar.gz
         ceb87a5f76b6352d28fd030aae5ad2165a133e9a8a6309891e793911203fc0ada9fb254dc05d183eaaa7e2b9851d3f1755b33f08fa6ff5b4b415ac4272bfe150 7790851 dovecot_2.3.19.1+dfsg1.orig.tar.gz
         2e5f62cb3d9685b57f76dbdf81edd2de6df7063ce96d39e75241ee63ab522acd7c7be932ed5d6ed898e5d6e179dd7e226e011226fd3b4271af562af8fe56ec89 67620 dovecot_2.3.19.1+dfsg1-2ubuntu4.debian.tar.xz
         """,  # noqa: E501
        'Testsuite-Triggers': 'lsb-release, python3, systemd-sysv',
        }

    source = SourcePackage(package_data)

    print('* Object:')
    pp.pprint(vars(source))

    print()
    print('* Provides Binaries:')
    for pkg in sorted(source.provides_binaries.keys()):
        print(f'  - {pkg}')

    print()
    print('* Build Dependencies:')
    for pkg, ver in sorted(source.build_dependencies.items()):
        print(f'  - {pkg}: {ver}')



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2023 Authors
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.
#
# Authors:
#   Bryce Harrington <bryce@canonical.com>

"""Interprets and analyzes an Ubuntu Apt record for a binary package."""

from functools import lru_cache
from typing import Dict

from .dict import unpack_to_dict


class BinaryPackage:
    """The Apt record about an installable binary .deb.

    SourcePackages are built into multiple BinaryPackages for different
    components and different architectures.

    This object is intended to be instantiated from a raw Apt binary
    package record such as extracted from a Packages.xz file by
    apt_pkg.TagFile().  The BinaryPackage class members handle the
    parsing and interpretation of the data values of the record such as
    installation dependencies.

    Since objects of this class are constructed as thin wrappers on the
    Apt record, the available member properties for the object will vary
    depending on the data elements present in the record.  For example,
    if there isn't a "Task:" field, then this object will not have a
    'self.task' property.

    For member properties needed by users of this class, it's
    recommended to guarantee their presence either by requiring them
    from the pkg_dict in the __init__() property, or add a property to
    the class that checks and substitutes a suitable default if missing.
    """
    # pylint: disable=no-member

    def __init__(self, pkg_dict: dict):
        """Initialize a new SourcePackage object for a given package.

        :param dict[str, str] pkg_dict: Data collection loaded from an Apt record.
        """
        # Auto-create member parameters from the Apt record's data structure
        for k, val in dict(pkg_dict).items():
            setattr(self, k.lower().replace('-', '_'), val)

        # Required fields that must be present in pkg_dict for object validity
        for field in ['package', 'version', 'architecture']:
            if getattr(self, field, None) is None:
                raise ValueError(f'undefined {field} from Apt record for binary package')

    def __repr__(self):
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'pkg_dict={vars(self)!r})')

    def __str__(self):
        """Return a human-readable textual description of the BinaryPackage.

        :rtype: str
        :returns: Human-readable string.
        """
        return f"{self.package} ({self.version}) [{self.architecture}]"

    @property
    def name(self):
        """The name of the binary package as recorded in the Apt database.

        :rtype: str
        :returns: Package name.
        """
        return self.package

    @property
    @lru_cache
    def installation_dependencies(self) -> Dict[str, str]:
        """Required binary packages that must be installed as prerequisites.

        :rtype: dict[str, str]
        :returns: Collection of package names to version specification strings.
        """
        if getattr(self, 'depends', None) is None:
            # Missing Depends is uncommon, but can occur for packages
            # consisting just of config files, for example.
            return {}
        deps = unpack_to_dict(self.depends, key_sep=' ')
        if not deps:
            raise RuntimeError('Could not parse packages from Depends line of Apt record.')
        return deps

    @property
    @lru_cache
    def recommended_packages(self) -> Dict[str, str]:
        """Optional binary packages intended to be co-installed.

        :rtype: dict[str, str]
        :returns: Collection of package names to version specification strings.
        """
        if getattr(self, 'recommends', None) is None:
            return {}
        recs = unpack_to_dict(self.recommends, key_sep=' ')
        if not recs:
            raise RuntimeError('Could not parse packages from Recommends line of Apt record.')
        return recs


if __name__ == "__main__":
    # pylint: disable=line-too-long, invalid-name

    import os
    import apt_pkg

    import pprint
    pp = pprint.PrettyPrinter(indent=4)

    root_dir = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
    tests_data_dir = os.path.join(root_dir, 'tests', 'data')
    assert os.path.exists(tests_data_dir), f'cannot find {tests_data_dir}'

    print('####################################')
    print('## BinaryPackage class smoke test ##')
    print('####################################')
    print()

    print('Binary Packages')
    print('---------------')
    with apt_pkg.TagFile(f'{tests_data_dir}/binary-amd64/Packages.xz') as pkgs:
        results = []
        for pkg in pkgs:
            binary = BinaryPackage(pkg)
            results.append(str(binary))

        num_results = len(results)
        ellipses_shown = False
        for i, result in enumerate(results):
            if i < 10 or i >= num_results - 10:
                print(result)
            elif not ellipses_shown:
                print('...')
                ellipses_shown = True
    print()

    print('Details for "dovecot-core"')
    print('--------------------------')
    package_data = {
        'Package': 'dovecot-core',
        'Architecture': 'amd64',
        'Version': '1:2.3.19.1+dfsg1-2ubuntu4',
        'Priority': 'optional',
        'Section': 'mail',
        'Source': 'dovecot',
        'Origin': 'Ubuntu',
        'Maintainer': 'Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>',
        'Original-Maintainer': 'Dovecot Maintainers <dovecot@packages.debian.org>',
        'Bugs': 'https://bugs.launchpad.net/ubuntu/+filebug',
        'Installed-Size': '10406',
        'Provides': 'dovecot-abi-2.3.abiv19, dovecot-common',
        'Pre-Depends': 'init-system-helpers (>= 1.54~)',
        'Depends': 'adduser, libpam-runtime, lsb-base, openssl, ssl-cert, ucf, libapparmor1 (>= 2.7.0~beta1+bzr1772), libbz2-1.0, libc6 (>= 2.36), libcap2 (>= 1:2.10), libcrypt1 (>= 1:4.1.0), libexttextcat-2.0-0 (>= 3.3.0), libicu72 (>= 72.1~rc-1~), liblua5.3-0, liblz4-1 (>= 0.0~r130), liblzma5 (>= 5.1.1alpha+20120614), libpam0g (>= 0.99.7.1), libsodium23 (>= 1.0.13), libssl3 (>= 3.0.0), libstemmer0d (>= 0+svn527), libsystemd0, libtirpc3 (>= 1.0.2), libunwind8, libwrap0 (>= 7.6-4~), libzstd1 (>= 1.5.2), zlib1g (>= 1:1.1.4)',  # noqa: E501
        'Suggests': 'dovecot-gssapi, dovecot-imapd, dovecot-ldap, dovecot-lmtpd, dovecot-lucene, dovecot-managesieved, dovecot-mysql, dovecot-pgsql, dovecot-pop3d, dovecot-sieve, dovecot-solr, dovecot-sqlite, dovecot-submissiond, ntp',  # noqa: E501
        'Breaks': 'dovecot-common (<< 1:2.0.14-2~), mailavenger (<< 0.8.1-4)',
        'Replaces': 'dovecot-common (<< 1:2.0.14-2~), mailavenger (<< 0.8.1-4)',
        'Filename': 'pool/main/d/dovecot/dovecot-core_2.3.19.1+dfsg1-2ubuntu4_amd64.deb',
        'Size': '3300962',
        'MD5sum': 'c6f61ffe0f01f51405c4fc22f6770cd2',
        'SHA1': 'ed389250d8738c0199f24cab2fe33e12b84e31c7',
        'SHA256': 'd73fb4bc55c764b9ad8c143c1f2b04a1cfb6ce2c2d751e4970ee09199ed7c3cb',
        'SHA512': 'ab55bdcdc31eac59168883ff83b2e6f2ee57b9340eb6b4e5b0f4e354813ae7dc0ca21ba34265f872a9c014b3c06bacad95708059c7a06c8c93d4a22ccb31494c',  # noqa: E501
        'Homepage': 'https://dovecot.org/',
        'Description': 'secure POP3/IMAP server - core files',
        'Task': 'mail-server',
        'Description-md5': '42825422b1ef9e3a592c94dfafed375c',
    }

    binary = BinaryPackage(package_data)

    print('* Object:')
    pp.pprint(vars(binary))

    print()
    print('* Installation Requires:')
    for pkg, ver in sorted(binary.installation_dependencies.items()):
        print(f'  - {pkg}: {ver}')

    print()
    print('* Recommends:')
    for pkg, ver in sorted(binary.recommended_packages.items()):
        print(f'  - {pkg}: {ver} ')



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the package's version tracking module.

Note this is for PpaDevTool's own versioning system.  This is unrelated
to Ubuntu or Debian version numbering.
"""

import os
import sys

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa._version import __version__, __version_info__


def test_version():
    """Checks that the __version__ is specified correctly."""
    assert type(__version__) is str
    assert '.' in __version__
    assert __version__[0].isdigit()
    assert __version__[-1] != '.'


def test_version_info():
    """Checks that the __version_info__ is specified correctly."""
    assert type(__version_info__) is tuple
    assert len(__version_info__) > 1
    for elem in __version_info__:
        assert type(elem) is int
        assert elem >= 0



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2021 Bryce W. Harrington
#
# Released under GNU AGPL or later, read the file 'LICENSE.AGPL' for
# more information.

# Extraction of bileto's lp class, for general use in other places

"""Launchpad Interface."""

from contextlib import suppress
from functools import lru_cache

from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import Credentials


class Lp:
    """High level wrapper object for Launchpad's API.

    This class wrappers the Launchpadlib service to cache object queries
    and to provide functionalies frequently needed when writing software
    for managing the Ubuntu distribution.

    This can be used as a drop-in replacement in scripts that already
    use Launchpadlib.  Simply replace your Launchpadlib.login_with() call
    with an instantiation of this class.  Any call that Lp does not handle
    itself is passed directly to the Launchpadlib object, so the entire
    API is available in exactly the same way.
    """
    # pylint: disable=invalid-name
    ROOT_URL = 'https://launchpad.net/'
    API_ROOT_URL = 'https://api.launchpad.net/devel/'
    BUGS_ROOT_URL = 'https://bugs.launchpad.net/'
    CODE_ROOT_URL = 'https://code.launchpad.net/'

    _real_instance = None

    def __init__(self, application_name, service=Launchpad,
                 staging=False, credentials=None, readonly=False):
        """Create a Launchpad service object.

        Authentication with Launchpad is done lazily, not at object
        initialization but at the point it first needs to actually use
        Launchpad functionality.  This permits adjustment of the
        object's credentials or other properties as needed.

        If the `$LP_CREDENTIAL` environment variable is defined, its
        contents will be loaded as the credentials to pass to the
        Credentials.from_string() function.  Stored credentials must
        be formatted according to the requirements of launchpadlib's
        Credentials class.  For more information on this class see:
            https://git.launchpad.net/launchpadlib/tree/src/launchpadlib/credentials.py

        :param str application_name: The text name of the software using
            this class.
        :param Launchpad service: The launchpadlib service class or
            object to wrapper.
        :param bool staging: When true, operate against a test instance
            of Launchpad instead of the real one.
        :param str credentials: (Optional) Formatted OAuth information
            to use when authenticating with Launchpad.  If not provided,
            will automatically login to Launchpad as needed.
        """
        self._app_name = application_name
        self._service = service
        self._credentials = credentials
        if staging:
            self._service_root = 'qastaging'
            self.ROOT_URL = 'https://qastaging.launchpad.net/'
            self.API_ROOT_URL = 'https://api.qastaging.launchpad.net/devel/'
            self.BUGS_ROOT_URL = 'https://bugs.qastaging.launchpad.net/'
            self.CODE_ROOT_URL = 'https://code.qastaging.launchpad.net/'
        else:
            self._service_root = 'production'
        if readonly:
            self._access_levels = ['READ_PUBLIC']
        else:
            self._access_levels = ['WRITE_PRIVATE']

    def _get_instance_from_creds(self) -> 'Launchpad | None':
        """
        Get an instance of _service using stored credentials if defined,
        else return None.

        For more information on Launchpad credentials-based authentication see
        https://help.launchpad.net/API/launchpadlib#Authenticated_access_for_website_integration

        :rtype: Launchpad | None
        :returns: Logged in Launchpad instance if credentials available,
            else None
        """
        if self._credentials:
            cred = Credentials.from_string(self._credentials)
            return self._service(
                cred, None, None,
                service_root=self._service_root,
                version='devel'
            )
        return None

    def _get_instance_from_login(self) -> 'Launchpad':
        """
        Prompts the user to authorize the login of a new credential
        or use the cached one if it is available and valid

        :rtype: launchpadlib.launchpad.Launchpad
        :returns: Logged in Launchpad instance
        """
        return self._service.login_with(
            application_name=self._app_name,
            service_root=self._service_root,
            allow_access_levels=self._access_levels,
            version='devel',  # Need devel for copyPackage.
        )

    @property
    def _instance(self):
        """Cache LP object."""
        if not self._real_instance:
            self._real_instance = (
                self._get_instance_from_creds() or
                self._get_instance_from_login()
            )
        return self._real_instance

    @property
    @lru_cache()
    def _api_root(self):
        """Identify the root URL of the launchpad API."""
        return self._instance.resource_type_link.split('#')[0]

    def __getattr__(self, attr):
        """Wrap launchpadlib so tightly you can't tell the difference."""
        assert not attr.startswith('_'), f"Can't getattr for {attr}"
        instance = super(Lp, self).__getattribute__('_instance')
        return getattr(instance, attr)

    @property
    @lru_cache()
    def ubuntu(self):
        """Shorthand for Ubuntu object.

        :rtype: distribution
        :returns: The distribution object for 'ubuntu'.
        """
        return self.distributions['ubuntu']

    @lru_cache()
    def ubuntu_active_series(self):
        """Identify currently supported Ubuntu series.

        This includes the series currently under development, but not
        ones which are experimental or obsolete.

        :rtype: list of distro_series
        :returns: All active Launchpad distro series for the Ubuntu project.
        """
        return [s for s in self.ubuntu.series if s.active]

    @lru_cache()
    def ubuntu_devel_series(self):
        """Identify the in-development Ubuntu series.

        This returns just the series currently under development or in a
        pre-release freeze.  While this is returned as a list for
        compatibility, there will typically be no more than one element
        in it.  However, note that this may be empty if used during the
        interim between a release and the opening of the next
        development series.

        :rtype: list of distro_series
        :returns: The in-development Launchpad distro series for the Ubuntu project.
        """
        devel_series = self.ubuntu.getDevelopmentSeries()
        if len(devel_series) > 0:
            return devel_series
        return [s for s in self.ubuntu.series if s.status == 'Pre-release Freeze']

    @lru_cache()
    def ubuntu_stable_series(self):
        """Identify current stable Ubuntu series.

        This includes the series considered stable currently.

        :rtype: distro_series
        :returns: All stable Launchpad distro series for the Ubuntu project.
        """
        return [s for s in self.ubuntu.series if s.status == 'Current Stable Release']

    @property
    @lru_cache()
    def debian(self):
        """Shorthand for Debian object.

        :rtype: distribution
        :returns: The distribution object for 'debian'.
        """
        return self.distributions['debian']

    @lru_cache()
    def debian_active_series(self):
        """Identify currently supported Debian series.

        :rtype: list of distro_series
        :returns: All active Launchpad distro series for the Debian project.
        """
        return [s for s in self.debian.series if s.active]

    @lru_cache()
    def debian_experimental_series(self):
        """Shorthand for Debian experimental series.

        :rtype: distro_series
        :returns: The Launchpad distro series for the Debian project.
        """
        return next(iter([s for s in self.debian.series if s.name == 'experimental']), None)

    @lru_cache()
    def get_teams(self, user):
        """Retrieve list of teams that user belongs to.

        :param str user: Name of the user to look up.
        :rtype: list(str)
        :returns: List of team names.
        """
        with suppress(KeyError, TypeError):
            return [
                team.self_link.partition('~')[-1].partition('/')[0]
                for team in self.people[user].memberships_details]

    def load(self, url):
        """Return a lp resource from a launchpad url.

        :param str url: The launchpad resource URL.
        :rtype: varies
        :returns: Launchpadlib object corresponding to given url.
        """
        return self._instance.load(url)



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""A person or team that owns one or more PPAs in Launchpad."""

from functools import lru_cache
from lazr.restfulclient.errors import BadRequest

from .ppa import Ppa
from .text import o2str


class PpaAlreadyExists(BaseException):
    """Exception indicating a PPA operation could not be performed."""

    def __init__(self, ppa_name, message=None):
        """Initialize the exception object.

        :param str ppa_name: The name of the pre-existing PPA.
        :param str message: An error message.
        """
        self.ppa_name = ppa_name
        self.message = message

    def __str__(self):
        """Return a human-readable error message.

        :rtype str:
        :return: Error message about the failure.
        """
        if self.message:
            return self.message
        elif self.ppa_name:
            return f"The PPA {self.ppa_name} already exists"


class PpaGroup:
    """Represents a person or team that owns one or more PPAs.

    This class provides a proxy object for interacting with collections
    of PPA.
    """

    def __init__(self, service, name):
        """Initialize a new PpaGroup object for a named person or team.

        :param launchpadlib.service service: The Launchpad service object.
        :param str name: Launchpad username for a person or team.
        """
        if not service:
            raise ValueError("undefined service.")
        if not name:
            raise ValueError("undefined name.")

        self.service = service
        self.name = name

    def __repr__(self):
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (f'{self.__class__.__name__}('
                f'service={self.service!r}, name={self.name!r})')

    def __str__(self):
        """Return a human-readable summary of the object.

        :rtype: str
        :returns: Printable summary of the object.
        """
        return 'tbd'

    @property
    @lru_cache
    def owner(self):
        """The person or team that owns this collection of PPAs.

        :rtype: launchpadlib.person
        :returns: Launchpad person object that owns this PPA.
        """
        return self.service.people[self.name]

    def create(self, ppa_name='ppa', ppa_description=None, private=False):
        """Register a new PPA with Launchpad.

        If a description is not provided a default one will be generated.

        :param str ppa_name: Name for the PPA to create.
        :param str ppa_description: Text description of the PPA.
        :param bool private: Limit access to PPA to only subscribed users and owner.
        :rtype: Ppa
        :returns: A Ppa object that describes the created PPA.

        :raises PpaAlreadyExists: Raised if a PPA by this name already exists in Launchpad.
        """
        ppa_settings = {
            'description': ppa_description,
            'displayname': ppa_name,
            'private': private
        }

        try:
            self.owner.createPPA(
                name=ppa_name,
                **ppa_settings
            )
            self.owner.lp_save()
        except BadRequest as e:
            if "You already have a PPA" in o2str(e.content):
                raise PpaAlreadyExists(ppa_name, e.content)
            else:
                raise e

        return Ppa(ppa_name, self.name, ppa_description, service=self.service)

    @property
    @lru_cache
    def ppas(self):
        """Generator to access the PPAs in this group.

        :rtype: Iterator[ppa.Ppa]
        :returns: Each PPA in the group as a ppa.Ppa object.
        """
        for lp_ppa in self.owner.ppas:
            if '-deletedppa' in lp_ppa.name:
                continue
            yield Ppa(lp_ppa.name, self.name,
                      service=self.service)

    @lru_cache
    def get(self, ppa_name):
        """Provide a Ppa for the named ppa.

        :rtype: ppa.Ppa
        :returns: A Ppa object describing the named ppa.
        """
        lp_ppa = self.owner.getPPAByName(name=ppa_name)
        if not lp_ppa:
            return None
        return Ppa(lp_ppa.name, self.name,
                   service=self.service)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2019 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

# pylint: disable=invalid-name

import os
import sys
from functools import lru_cache
from unittest.mock import Mock

from launchpadlib.credentials import Credentials

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.ppa import Ppa
from ppa.constants import ARCHES_PPA_DEFAULT
from ppa.ppa_group import PpaAlreadyExists


class SeriesMock:
    def __init__(self, name):
        self.name = name


class PublicationMock:
    def __init__(self, name, version, status, series):
        self.source_package_name = name
        self.source_package_version = version
        self.status = status
        self.distro_series = SeriesMock(series)


class ProcessorMock:
    """A stand-in for a Launchpad Processor object."""
    def __init__(self, name):
        self.name = name


class ArchiveDependencyMock:
    """A stand-in for a Launchpad ArchiveDependency object."""
    def __init__(self, archive, dependency, pocket, component):
        self.archive = archive
        self.dependency = dependency
        self.pocket = pocket
        self.component = component


class ArchiveMock:
    """A stand-in for a Launchpad Archive object."""
    def __init__(self, name, description, owner, distribution):
        self.displayname = name
        self.description = description
        self.owner = owner
        self.private = False
        self.processors = [ProcessorMock(proc_name) for proc_name in ARCHES_PPA_DEFAULT]
        self.publish = True
        self.published_sources = []
        self.published_binaries = []
        self.distribution = distribution
        self.dependencies = []
        self.web_link = None

    def setProcessors(self, processors):
        self.processors = [ProcessorMock(proc.split('/')[-1]) for proc in processors]

    def getPublishedSources(self, **params):
        return self.published_sources

    def getPublishedBinaries(self, **params):
        return self.published_binaries

    def lp_save(self):
        return True

    def addArchiveDependency(self, component=None, dependency=None, pocket=None):
        self.dependencies.append(ArchiveDependencyMock(self, dependency, pocket, component))


class PersonMock:
    """A stand-in for a Launchpad Person object."""
    def __init__(self, name):
        self.name = name
        self._ppas = []

    def createPPA(self, name, description, displayname, private=None):
        for ppa in self._ppas:
            if ppa.name == name:
                raise PpaAlreadyExists(name)
        ppa = Ppa(name, self.name, description)
        Ppa.archive = ArchiveMock(ppa.name, ppa.description, self, DEFAULT_DISTRO)
        if isinstance(private, bool):
            Ppa.archive.private = private
        self._ppas.append(ppa)
        return True

    def getPPAByName(self, name):
        for ppa in self._ppas:
            if ppa.name == name:
                return ppa.archive
        return None

    def lp_save(self):
        return True

    @property
    def ppas(self):
        return self._ppas


class DistributionMock:
    def __init__(self, name):
        self.name = name
        self.main_archive = ArchiveMock(
                "primary",
                "primary archive",
                PersonMock(f"{name}-archive"), self
        )

    def getSeries(self, name_or_version):
        return SeriesMock(name_or_version)


DEFAULT_DISTRO = DistributionMock('mydistro')


class LaunchpadMock:
    """A stand-in for Launchpad."""
    def __init__(self):
        self.people = {'me': PersonMock('me')}

    def add_person(self, name):
        print(f"Adding person {name}")
        self.people[name] = PersonMock(name)

    @property
    def me(self):
        return self.people['me']


class LpServiceMock:
    """A stand-in for the Lp service object."""
    ROOT_URL = 'https://mocklaunchpad.net/'
    API_ROOT_URL = 'https://api.mocklaunchpad.net/devel/'
    BUGS_ROOT_URL = 'https://bugs.mocklaunchpad.net/'
    CODE_ROOT_URL = 'https://code.mocklaunchpad.net/'

    def __init__(self, credentials=None):
        self.launchpad = LaunchpadMock()
        self._credentials = credentials

    @property
    @lru_cache
    def credentials(self):
        return Mock(Credentials)

    @property
    def me(self):
        return self.launchpad.people['me']

    @property
    def people(self):
        return self.launchpad.people

    def get_bug(self, bug_id):
        class BugMock:
            @property
            def title(self):
                return "Mock bug report"

            @property
            def description(self):
                return "Description line 1\n\ndescription line 2"

        return BugMock()


class RequestResponseMock:
    """A stand-in for a request result."""
    def __init__(self, text):
        self._text = text.encode('utf-8')

    def read(self):
        """Simply returns the exact text provided in initializer."""
        return self._text



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2024 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""A package release for a particular distro series."""

from functools import lru_cache

from .lp import Lp
from .constants import URL_AUTOPKGTEST


class SourcePublication:
    """Encapsulate data needed to represent a source publication record.

    This class can be used to hold the data 'coordinates' for a source
    package release, namely the package's name, version, distro series,
    and the archive (PPA) containing it.
    """

    def __init__(self, package, version, series, archive=None, status=None):
        """Initialize a new SourcePublication object for a given SPPH.

        This creates only the local representation of the SPPH, it does
        not cause a new SPPH to be created in or loaded from Launchpad.

        :param str package: The source package name.
        :param str version: The version of the source package to install.
        :param str series: The distro release series codename.
        :param Ppa archive: The Ppa owning this source publication, or
            None if it belongs to the primary archive.
        :param str status: The Launchpad publication state (Pending,
            Published, Superseded, Deleted, Obsolete)
        """
        if not package:
            raise ValueError('undefined package.')
        if not version:
            raise ValueError('undefined version.')
        if not series:
            raise ValueError('undefined series.')

        self.package = package
        self.version = version
        self.series = series
        self.archive = archive
        self.status = status

    def __repr__(self) -> str:
        """Return a machine-parsable unique representation of object.

        :rtype: str
        :returns: Official string representation of the object.
        """
        return (
            f'{self.__class__.__name__}('
            f'package={self.package!r}, '
            f'version={self.version!r}, '
            f'series={self.series!r}, '
            f'archive={self.archive!r}, '
            f'status={self.status!r})'
        )

    def __str__(self) -> str:
        """Return a displayable string identifying the SPPH.

        :rtype: str
        :returns: Human-readable string.
        """
        address = f"{self.series}/{self.package}/{self.version}"
        if self.archive:
            address += f" ({self.archive})"
        if self.status:
            address += f": {self.status}"
        return address

    @lru_cache
    def to_dict(self) -> dict:
        """Return a basic dict structure of the SourcePublication's data."""
        return {
            'package': self.package,
            'version': self.version,
            'series': self.series,
            'archive': self.archive,
            'status': self.status
        }

    @property
    @lru_cache
    def autopkgtest_history_url(self) -> str:
        """Renders the SPPH as a URL to the autopkgtest history.

        :rtype: str
        :returns: Autopkgtest history URL.
        """
        if self.package.startswith('lib'):
            prefix = self.package[0:4]
        else:
            prefix = self.package[0]
        pkg_str = f"{prefix}/{self.package}"
        return f"{URL_AUTOPKGTEST}/packages/{pkg_str}/{self.series}"


if __name__ == "__main__":
    print('##################################')
    print('## Publication class smoke test ##')
    print('##################################')
    print()

    lp = Lp('smoketest', staging=True)
    pub = SourcePublication('p', '1.0', 'noble', None, 'Published')
    print(repr(pub))
    print()
    print(str(pub))
    print()
    print(pub.to_dict())
    print()
    print(pub.autopkgtest_history_url)



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Author:  Bryce Harrington <bryce@canonical.com>
#
# Copyright (C) 2022 Bryce W. Harrington
#
# Released under GNU GPLv2 or later, read the file 'LICENSE.GPLv2+' for
# more information.

"""Tests the Subtest class as representing components of an autopkgtest run."""

import os
import sys
import json

import pytest

sys.path.insert(0, os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")))

from ppa.subtest import Subtest


def test_object():
    """Checks that Subtest objects can be instantiated."""
    subtest = Subtest('a UNKNOWN')
    # TODO: If no ':' in description, should throw exception?
    # TODO: Or add a function that parses lines and returns subtests?
    assert subtest


def test_repr():
    """Checks Subtest object representation."""
    subtest = Subtest('a PASS')
    assert repr(subtest) == "Subtest(line='a PASS')"


def test_str():
    """Checks Subtest object textual presentation."""
    subtest = Subtest('a PASS')
    assert 'PASS' in f"{subtest}"
    assert 'a' in f"{subtest}"
    assert '🟩' in f"{subtest}"


def test_to_dict():
    """Checks Subtest object structural representation."""
    subtest = Subtest('a PASS')
    expected_keys = ['desc', 'line', 'status', 'status_icon']
    expected_types = [str]

    d = subtest.to_dict()
    assert isinstance(d, dict), f"type of d is {type(d)} not dict"

    # Verify expected keys are present
    assert sorted(d.keys()) == sorted(expected_keys)

    # Verify values are within set of expected types
    for k, v in d.items():
        assert type(v) in expected_types, f"'{k}={v}' is unexpected type {type(v)}"

    # Verify full dict can be written as JSON
    try:
        assert json.dumps(d)
    except UnicodeDecodeError as e:
        assert False, f"Wrong UTF codec detected: {e}"
    except json.JSONDecodeError as e:
        assert False, f"JSON decoding error: {e.msg}, {e.doc}, {e.pos}"


def test_desc():
    """Checks Subtest description is parsed correctly."""
    subtest = Subtest('a PASS')
    assert subtest.desc == 'a'

    subtest = Subtest('a FAIL descriptive text')
    assert subtest.desc == 'a'

    subtest = Subtest('a:b  PASS')
    assert subtest.desc == 'a:b'


@pytest.mark.parametrize('line, status', [
    ('a PASS', 'PASS'),
    ('b SKIP', 'SKIP'),
    ('c FAIL', 'FAIL'),
    ('d FAIL non-zero exit status 123', 'FAIL'),
    ('librust-clang-sys-dev:clang_10_0 FAIL non-zero exit status 101', 'FAIL'),
    ('librust-clang-sys-dev:static FLAKY non-zero exit status 101', 'FLAKY'),
    ('f BAD', 'BAD'),
    ('g UNKNOWN', 'UNKNOWN'),
    ('h invalid', 'UNKNOWN'),
    ('i bAd', 'UNKNOWN'),
])
def test_status(line, status):
    """Checks Subtest status is parsed correctly."""
    subtest = Subtest(line)
    assert subtest.status == status

    subtest = Subtest('a PASS')


@pytest.mark.parametrize('line, icon', [
    ('x PASS', "🟩"),
    ('x SKIP', "🟧"),
    ('x FAIL', "🟥"),
    ('x FLAKY', "🟫"),
    ('x BAD', "⛔"),
    ('x UNKNOWN', "⚪"),
    ('x invalid', "⚪"),
    ('x bAd', "⚪"),
])
def test_status_icon(line, icon):
    """Checks Subtest provides correct icon for status.

    :param str line: Subtest status line to be parsed.
    :param str icon: Resulting status icon that should be returned.
    """
    subtest = Subtest(line)
    assert subtest.status_icon == icon
